from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    Iterable,
    TypeVar,
    cast,
    overload,
)

from ._base import BaseDoc, BaseEvent, BaseType, Typed, base_types, event_types
from ._pycrdt import Map as _Map
from ._pycrdt import MapEvent as _MapEvent
from ._pycrdt import Subscription

if TYPE_CHECKING:
    from ._doc import Doc

T = TypeVar("T")
T_DefaultValue = TypeVar("T_DefaultValue")


class Map(BaseType, Generic[T]):
    """
    A collection used to store key-value entries in an unordered manner, similar to a Python `dict`.
    """

    _prelim: dict | None
    _integrated: _Map | None

    def __init__(
        self,
        init: dict[str, T] | None = None,
        *,
        _doc: Doc | None = None,
        _integrated: _Map | None = None,
    ) -> None:
        """
        Creates a map with an optional initial value:
        ```py
        map0 = Map()
        map1 = Map({"foo": 0, "bar": 3, "baz": map0})
        ```

        Args:
            init: The list from which to initialize the array.
        """
        super().__init__(
            init=init,
            _doc=_doc,
            _integrated=_integrated,
        )

    def _init(self, value: dict[str, T] | None) -> None:
        if value is None:
            return
        with self.doc.transaction():
            for k, v in value.items():
                self._set(k, v)

    def _set(self, key: str, value: T) -> None:
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(value, BaseDoc):
                # subdoc
                self.integrated.insert_doc(txn._txn, key, value._doc)
            elif isinstance(value, BaseType):
                # shared type
                assert txn._txn is not None
                self._do_and_integrate("insert", value, txn._txn, key)
            else:
                # primitive type
                self.integrated.insert(txn._txn, key, value)

    def _get_or_insert(self, name: str, doc: Doc) -> _Map:
        assert doc._txn is not None
        assert doc._txn._txn is not None
        return doc._doc.get_or_insert_map(doc._txn._txn, name)

    def __len__(self) -> int:
        """
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 1})
        assert len(map0) == 2
        ```
        Returns:
            The length of the map.
        """
        with self.doc.transaction() as txn:
            return self.integrated.len(txn._txn)

    def __str__(self) -> str:
        """
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 1})
        assert str(map0) == '{"foo":0,"bar":1}'
        ```

        Returns:
            The string representation of the map.
        """
        with self.doc.transaction() as txn:
            return self.integrated.to_json(txn._txn)

    def to_py(self) -> dict[str, T] | None:
        """
        Recursively converts the map's items to Python objects, and
        returns them in a `dict`. If the map was not yet inserted in a document,
        returns `None` if the map was not initialized.

        Returns:
            The map recursively converted to Python objects, or `None`.
        """
        if self._integrated is None:
            py = self._prelim
            if py is None:
                return None
        else:
            py = dict(self)
        for key, val in py.items():
            if isinstance(val, BaseType):
                py[key] = val.to_py()
        return py

    def __delitem__(self, key: str) -> None:
        """
        Removes the item at the given key from the map:
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 1})
        del map0["foo"]
        assert map0.to_py() == {"bar": 1}
        ```

        Args:
            key: The key of the item to remove.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            self._check_key(key)
            self.integrated.remove(txn._txn, key)

    def __getitem__(self, key: str) -> T:
        """
        Gets the value at the given key:
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 1})
        assert map0["foo"] == 0
        ```

        Returns:
            The value at the given key.
        """
        with self.doc.transaction() as txn:
            self._check_key(key)
            return self._maybe_as_type_or_doc(self.integrated.get(txn._txn, key))

    def __setitem__(self, key: str, value: T) -> None:
        """
        Sets a value at the given key:
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 1})
        map0["foo"] = 2
        assert map0["foo"] == 2
        ```

        Args:
            key: The key to set.
            value: The value to set.

        Raises:
            RuntimeError: Key must be of type string.
        """
        if not isinstance(key, str):
            raise RuntimeError("Key must be of type string")
        with self.doc.transaction():
            self._set(key, value)

    def __iter__(self) -> Iterable[str]:
        """
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 0})
        for key in m0:
            assert map[key] == 0
        ```
        Returns:
            An iterable over the keys of the map.
        """
        return self.keys()

    def __contains__(self, item: str) -> bool:
        """
        Checks if the given key is in the map:
        ```py
        Doc()["map0"] = map0 = Map({"foo": 0, "bar": 0})
        assert "baz" not in map0:
        ```

        Args:
            item: The key to look for in the map.

        Returns:
            True if the key was found.
        """
        # use integrated.has to avoid fetching all keys
        with self.doc.transaction() as txn:
            return self.integrated.has(txn._txn, item)

    @overload
    def get(self, key: str) -> T | None: ...

    @overload
    def get(self, key: str, default_value: T_DefaultValue) -> T | T_DefaultValue: ...

    def get(self, *args):
        """
        Returns the value corresponding to the given key if it exists, otherwise
        returns the default value if passed, or `None`.

        Args:
            args: The key of the value to get, and an optional default value.

        Returns:
            The value at the given key, or the default value or `None`.
        """
        key, *default_value = args
        with self.doc.transaction():
            if self.__contains__(key):
                return self[key]
            if not default_value:
                return None
            return default_value[0]

    @overload
    def pop(self, key: str) -> T: ...

    @overload
    def pop(self, key: str, default_value: T_DefaultValue) -> T | T_DefaultValue: ...

    def pop(self, *args):
        """
        Removes the entry at the given key from the map, and returns the corresponding value.

        Args:
            args: The key of the value to pop, and an optional default value.

        Returns:
            The value at the given key, or the default value if passed.
        """
        key, *default_value = args
        with self.doc.transaction():
            if not self.__contains__(key):
                if not default_value:
                    raise KeyError
                return default_value[0]
            res = self[key]
            if isinstance(res, BaseType):
                res = res.to_py()
            del self[key]
            return res

    def _check_key(self, key: str) -> None:
        if not isinstance(key, str):
            raise RuntimeError("Key must be of type string")
        if not self.__contains__(key):
            raise KeyError(key)

    def keys(self) -> Iterable[str]:
        """
        Returns:
            An iterable over the keys of the map.
        """
        with self.doc.transaction() as txn:
            return iter(self.integrated.keys(txn._txn))

    def values(self) -> Iterable[T]:
        """
        Returns:
            An iterable over the values of the map.
        """
        with self.doc.transaction() as txn:
            for k in self.integrated.keys(txn._txn):
                yield self[k]

    def items(self) -> Iterable[tuple[str, T]]:
        """
        Returns:
            An iterable over the key-value pairs of the map.
        """
        with self.doc.transaction() as txn:
            for k in self.integrated.keys(txn._txn):
                yield k, self[k]

    def clear(self) -> None:
        """
        Removes all entries from the map.
        """
        with self.doc.transaction() as txn:
            for k in self.integrated.keys(txn._txn):
                del self[k]

    def update(self, value: dict[str, T]) -> None:
        """
        Sets entries in the map from all entries in the passed `dict`.

        Args:
            value: The `dict` from which to get the entries to update.
        """
        self._init(value)

    def observe(self, callback: Callable[[MapEvent], None]) -> Subscription:
        """
        Subscribes a callback to be called with the map event.

        Args:
            callback: The callback to call with the [MapEvent][pycrdt.MapEvent].
        """
        return super().observe(cast(Callable[[BaseEvent], None], callback))


class TypedMap(Typed):
    """
    A container for a [Map][pycrdt.Map.__init__] where values have types associated with
    specific keys. The underlying `Map` can be accessed with the special `_` attribute.

    ```py
    from pycrdt import Array, Map, TypedDoc, TypedMap

    class MyMap(TypedMap):
        name: str
        toggle: bool
        nested: Array[bool]

    class MyDoc(TypedDoc):
        map0: MyMap

    doc = MyDoc()

    doc.map0.name = "John"
    doc.map0.toggle = True
    doc.map0.nested = Array([True, False])

    print(doc.map0._.to_py())
    # {'nested': [True, False], 'toggle': True, 'name': 'John'}
    ```
    """

    _: Map

    def __init__(self, map: TypedMap | Map | None = None) -> None:
        super().__init__()
        if map is None:
            map = Map()
        elif isinstance(map, TypedMap):
            map = map._
        self._ = map


class MapEvent(BaseEvent):
    """
    A map change event.

    Attributes:
        target (Map): The changed map.
        delta (list[dict[str, Any]]): A list of items describing the changes.
        path (list[int | str]): A list with the indices pointing to the map that was changed.
        transaction (ReadTransaction): The transaction this change was done in.
    """

    __slots__ = "target", "keys", "path", "transaction"


base_types[_Map] = Map
event_types[_MapEvent] = MapEvent



from __future__ import annotations

import threading
from abc import ABC, abstractmethod
from functools import partial
from inspect import signature
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Type,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

import anyio
from anyio import BrokenResourceError, create_memory_object_stream
from anyio.abc import TaskGroup
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

from ._pycrdt import Doc as _Doc
from ._pycrdt import Subscription
from ._pycrdt import Transaction as _Transaction
from ._sticky_index import Assoc, StickyIndex
from ._transaction import ReadTransaction, Transaction

if TYPE_CHECKING:
    from ._doc import Doc


base_types: dict[Any, type[BaseType | BaseDoc]] = {}
event_types: dict[Any, type[BaseEvent]] = {}


def forbid_read_transaction(txn: Transaction):
    if isinstance(txn, ReadTransaction):
        raise RuntimeError("Read-only transaction cannot be used to modify document structure")


class BaseDoc:
    _doc: _Doc
    _twin_doc: BaseDoc | None
    _txn: Transaction | None
    _exceptions: list[Exception]
    _txn_lock: threading.Lock
    _txn_async_lock: anyio.Lock
    _allow_multithreading: bool
    _Model: Any
    _subscriptions: list[Subscription]
    _origins: dict[int, Any]
    _task_group: TaskGroup | None

    def __init__(
        self,
        *,
        client_id: int | None = None,
        skip_gc: bool | None = None,
        doc: _Doc | None = None,
        Model=None,
        allow_multithreading: bool = False,
        **data,
    ) -> None:
        super().__init__(**data)
        if doc is None:
            doc = _Doc(client_id, skip_gc)
        self._doc = doc
        self._txn = None
        self._exceptions = []
        self._txn_lock = threading.Lock()
        self._txn_async_lock = anyio.Lock()
        self._Model = Model
        self._subscriptions = []
        self._origins = {}
        self._allow_multithreading = allow_multithreading
        self._task_group = None


class BaseType(ABC):
    _doc: Doc | None
    _prelim: Any | None
    _integrated: Any
    _type_name: str
    _subscriptions: list[Subscription]

    def __init__(
        self,
        init: Any = None,
        *,
        _doc: Doc | None = None,
        _integrated: Any = None,
    ) -> None:
        self._type_name = self.__class__.__name__.lower()
        self._subscriptions = []
        self._send_streams: dict[bool, set[MemoryObjectSendStream[BaseEvent | list[BaseEvent]]]] = {
            False: set(),
            True: set(),
        }
        self._event_subscription: dict[bool, Subscription] = {}
        # private API
        if _integrated is not None:
            self._doc = _doc
            self._prelim = None
            self._integrated = _integrated
            return
        # public API
        self._doc = None
        self._prelim = init
        self._integrated = None

    @abstractmethod
    def to_py(self) -> Any: ...

    @abstractmethod
    def _get_or_insert(self, name: str, doc: Doc) -> Any: ...

    @abstractmethod
    def _init(self, value: Any | None) -> None: ...

    def _forbid_read_transaction(self, txn: Transaction):
        forbid_read_transaction(txn)

    def _integrate(self, doc: Doc, integrated: Any) -> Any:
        prelim = self._prelim
        self._doc = doc
        self._prelim = None
        self._integrated = integrated
        return prelim

    def _do_and_integrate(self, action: str, value: BaseType, txn: _Transaction, *args) -> None:
        method = getattr(self._integrated, f"{action}_{value.type_name}_prelim")
        integrated = method(txn, *args)
        assert self._doc is not None
        prelim = value._integrate(self._doc, integrated)
        value._init(prelim)

    def _maybe_as_type_or_doc(self, obj: Any) -> Any:
        for k, v in base_types.items():
            if isinstance(obj, k):
                if issubclass(v, BaseDoc):
                    # create a BaseDoc
                    return v(doc=obj)
                # create a BaseType
                return v(_doc=self.doc, _integrated=obj)
        # that was a primitive value, just return it
        return obj

    @property
    def integrated(self) -> Any:
        if self._integrated is None:
            raise RuntimeError("Not integrated in a document yet")
        return self._integrated

    @property
    def doc(self) -> Doc:
        """
        The document this shared type belongs to.

        Raises:
            RuntimeError: Not integrated in a document yet.
        """
        if self._doc is None:
            raise RuntimeError("Not integrated in a document yet")
        return self._doc

    @property
    def is_prelim(self) -> bool:
        return self._prelim is not None

    @property
    def is_integrated(self) -> bool:
        return self._integrated is not None

    @property
    def prelim(self) -> Any:
        return self._prelim

    @property
    def type_name(self) -> str:
        return self._type_name

    def observe(self, callback: Callable[[BaseEvent], None]) -> Subscription:
        param_nb = len(signature(callback).parameters)
        _callback = partial(observe_callback, callback, self.doc, param_nb)
        subscription = self.integrated.observe(_callback)
        self._subscriptions.append(subscription)
        return subscription

    def observe_deep(self, callback: Callable[[list[BaseEvent]], None]) -> Subscription:
        """
        Subscribes a callback for all events emitted by this and nested collaborative types.

        Args:
            callback: The callback to call with the list of events.
        """
        param_nb = len(signature(callback).parameters)
        _callback = partial(observe_deep_callback, callback, self.doc, param_nb)
        subscription = self.integrated.observe_deep(_callback)
        self._subscriptions.append(subscription)
        return subscription

    def unobserve(self, subscription: Subscription) -> None:
        """
        Unsubscribes to changes using the given subscription.

        Args:
            subscription: The subscription to unregister.
        """
        self._subscriptions.remove(subscription)
        subscription.drop()

    @overload
    def events(
        self,
        deep: Literal[False],
        max_buffer_size: float = float("inf"),
    ) -> MemoryObjectReceiveStream[BaseEvent]: ...

    @overload
    def events(
        self,
        deep: Literal[True],
        max_buffer_size: float = float("inf"),
    ) -> MemoryObjectReceiveStream[list[BaseEvent]]: ...

    def events(
        self,
        deep: bool = False,
        max_buffer_size: float = float("inf"),
    ):
        """
        Allows to asynchronously iterate over the shared type events, without using a callback.
        A buffer is used to store the events, allowing to iterate at a (temporarily) slower rate
        than they are produced.

        This method must be used with an async context manager and an async for-loop:

        ```py
        async def main():
            async with doc.events() as events:
                async for event in events:
                    update: bytes = event.update
                    ...
        ```

        Args:
            deep: Whether to iterate over the nested events.
            max_buffer_size: Maximum number of events that can be buffered.

        Returns:
            An async iterator over the shared type events.
        """
        observe = self.observe_deep if deep else self.observe
        if not self._send_streams[deep]:
            self._event_subscription[deep] = observe(partial(self._send_event, deep))
        send_stream, receive_stream = create_memory_object_stream[
            Union[BaseEvent, list[BaseEvent]]
        ](max_buffer_size=max_buffer_size)
        self._send_streams[deep].add(send_stream)
        return receive_stream

    def _send_event(self, deep: bool, event: BaseEvent | list[BaseEvent]):
        to_remove: list[MemoryObjectSendStream[BaseEvent | list[BaseEvent]]] = []
        send_streams = self._send_streams[deep]
        for send_stream in send_streams:
            try:
                send_stream.send_nowait(event)
            except BrokenResourceError:
                to_remove.append(send_stream)
        for send_stream in to_remove:
            send_stream.close()
            send_streams.remove(send_stream)
        if not send_streams:
            self.unobserve(self._event_subscription[deep])


class Sequence(BaseType):
    def sticky_index(self, index: int, assoc: Assoc = Assoc.AFTER) -> StickyIndex:
        """
        A permanent position that sticks to the same place even when
        concurrent updates are made.

        Args:
            index: The index at which to stick.
            assoc: The [Assoc][pycrdt.Assoc] specifying whether to stick to the location
                before or after the index.

        Returns:
            A [StickyIndex][pycrdt.StickyIndex] that can be used to retrieve the index after
            an update was applied.
        """
        return StickyIndex.new(self, index, assoc)


def observe_callback(
    callback: Callable[[], None] | Callable[[Any], None] | Callable[[Any, ReadTransaction], None],
    doc: Doc,
    param_nb: int,
    event: Any,
):
    with doc._read_transaction(event.transaction) as txn:
        _event = event_types[type(event)](event, doc)
        _event.transaction = txn
        params = (_event, txn)
        try:
            callback(*params[:param_nb])  # type: ignore[arg-type]
        except Exception as exc:
            doc._exceptions.append(exc)


def observe_deep_callback(
    callback: Callable[[], None] | Callable[[Any], None] | Callable[[Any, ReadTransaction], None],
    doc: Doc,
    param_nb: int,
    events: list[Any],
):
    with doc._read_transaction(events[0].transaction) as txn:
        for idx, event in enumerate(events):
            _event = event_types[type(event)](event, doc)
            _event.transaction = txn
            events[idx] = _event
        params = (events, txn)
        try:
            callback(*params[:param_nb])  # type: ignore[arg-type]
        except Exception as exc:
            doc._exceptions.append(exc)


class BaseEvent:
    transaction: ReadTransaction
    __slots__ = ("transaction",)

    def __init__(self, event: Any, doc: Doc):
        slot: str
        for slot in self.__slots__:
            processed = process_event(getattr(event, slot), doc)
            setattr(self, slot, processed)

    def __str__(self):
        str_list = []
        slot: Any
        for slot in self.__slots__:
            val = str(getattr(self, slot))
            str_list.append(f"{slot}: {val}")
        ret = ", ".join(str_list)
        return "{" + ret + "}"


def process_event(value: Any, doc: Doc) -> Any:
    if isinstance(value, list):
        for idx, val in enumerate(value):
            value[idx] = process_event(val, doc)
    elif isinstance(value, dict):
        for key, val in value.items():
            value[key] = process_event(val, doc)
    else:
        val_type = type(value)
        if val_type in base_types:
            if val_type is _Doc:
                doc_type: type[BaseDoc] = cast(Type[BaseDoc], base_types[val_type])
                value = doc_type(doc=value)
            else:
                base_type = cast(Type[BaseType], base_types[val_type])
                value = base_type(_integrated=value, _doc=doc)
    return value


class Typed:
    _: Any

    def __init__(self) -> None:
        self.__dict__["annotations"] = {
            name: _type
            for name, _type in get_type_hints(type(self).mro()[0]).items()
            if name != "_"
        }

    if not TYPE_CHECKING:

        def __getattr__(self, key: str) -> Any:
            annotations = self.__dict__["annotations"]
            if key not in annotations:
                raise AttributeError(f'"{type(self).mro()[0]}" has no attribute "{key}"')
            expected_type = annotations[key]
            if hasattr(expected_type, "mro") and Typed in expected_type.mro():
                return expected_type(self._[key])
            return self._[key]

        def __setattr__(self, key: str, value: Any) -> None:
            if key == "_":
                self.__dict__["_"] = value
                return
            annotations = self.__dict__["annotations"]
            if key not in annotations:
                raise AttributeError(f'"{type(self).mro()[0]}" has no attribute "{key}"')
            expected_type = annotations[key]
            origin = get_origin(expected_type)
            if origin in (Union, UnionType):
                expected_types = get_args(expected_type)
            elif origin is not None:
                expected_type = origin
                expected_types = (expected_type,)
            else:
                expected_types = (expected_type,)
            if type(value) not in expected_types:
                raise TypeError(
                    f'Incompatible types in assignment (expression has type "{expected_type}", '
                    f'variable has type "{type(value)}")'
                )
            if isinstance(value, Typed):
                value = value._
            self._[key] = value



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager
from logging import Logger, getLogger
from typing import AsyncIterator, Protocol

from anyio import (
    TASK_STATUS_IGNORED,
    Event,
    Lock,
    create_task_group,
)
from anyio.abc import TaskGroup, TaskStatus

from ._doc import Doc
from ._sync import (
    YMessageType,
    YSyncMessageType,
    create_sync_message,
    create_update_message,
    handle_sync_message,
)


class Channel(Protocol):
    """A transport-agnostic stream used to synchronize a document through a provider.
    An example of a channel is a WebSocket.

    Messages can be received through the channel using an async iterator,
    until the connection is closed:
    ```py
    async for message in channel:
        ...
    ```
    Or directly by calling `recv()`:
    ```py
    message = await channel.recv()
    ```
    Sending messages is done with `send()`:
    ```py
    await channel.send(message)
    ```
    """

    @property
    def path(self) -> str:
        """The channel path."""
        ...  # pragma: nocover

    def __aiter__(self) -> "Channel":
        return self

    async def __anext__(self) -> bytes:
        return await self.recv()

    async def send(self, message: bytes) -> None:
        """Send a message.

        Args:
            message: The message to send.
        """
        ...  # pragma: nocover

    async def recv(self) -> bytes:
        """Receive a message.

        Returns:
            The received message.
        """
        ...  # pragma: nocover


class Provider:
    def __init__(self, doc: Doc, channel: Channel, log: Logger | None = None) -> None:
        """A provider synchronizes a document through a channel.

        The provider should preferably be used with an async context manager:
        ```py
        async with provider:
            ...
        ```
        However, a lower-level API can also be used:
        ```py
        task = asyncio.create_task(provider.start())
        await provider.started.wait()
        ...
        await provider.stop()
        ```

        Arguments:
            doc: The `Doc` to connect through the `Channel`.
            channel: The `Channel` through which to connect the `Doc`.
            log: An optional logger.
        """
        self._doc = doc
        self._channel = channel
        self.log = log or getLogger(__name__)
        self.started = Event()
        self._start_lock = Lock()
        self._task_group: TaskGroup | None = None

    async def _run(self):
        sync_message = create_sync_message(self._doc)
        self.log.debug(
            "Sending %s message to endpoint: %s",
            YSyncMessageType.SYNC_STEP1.name,
            self._channel.path,
        )
        await self._channel.send(sync_message)
        assert self._task_group is not None
        self._task_group.start_soon(self._send_updates)
        async for message in self._channel:
            if message[0] == YMessageType.SYNC:
                self.log.debug(
                    "Received %s message from endpoint: %s",
                    YSyncMessageType(message[1]).name,
                    self._channel.path,
                )
                reply = handle_sync_message(message[1:], self._doc)
                if reply is not None:
                    self.log.debug(
                        "Sending %s message to endpoint: %s",
                        YSyncMessageType.SYNC_STEP2.name,
                        self._channel.path,
                    )
                    await self._channel.send(reply)

    async def _send_updates(self):
        async with self._doc.events() as events:
            async for event in events:
                message = create_update_message(event.update)
                await self._channel.send(message)

    async def __aenter__(self) -> Provider:
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(
                self._get_or_create_task_group()
            )
            await self._task_group.start(self.start)
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        await self.stop()
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    @asynccontextmanager
    async def _get_or_create_task_group(self) -> AsyncIterator[TaskGroup]:
        if self._task_group is not None:
            yield self._task_group
            return

        async with create_task_group() as tg:
            yield tg

    async def start(
        self,
        *,
        task_status: TaskStatus[None] = TASK_STATUS_IGNORED,
    ) -> None:
        """Start the provider.

        Args:
            task_status: The status to set when the task has started.
        """
        async with self._start_lock:
            async with self._get_or_create_task_group() as self._task_group:
                task_status.started()
                self.started.set()
                self._task_group.start_soon(self._run)

    async def stop(self) -> None:
        """Stop the provider."""
        assert self._task_group is not None
        self._task_group.cancel_scope.cancel()
        self._task_group = None



from __future__ import annotations

from functools import partial
from inspect import iscoroutinefunction
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Literal,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from anyio import BrokenResourceError, create_memory_object_stream
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

from ._base import BaseDoc, BaseType, Typed, base_types, forbid_read_transaction
from ._pycrdt import Doc as _Doc
from ._pycrdt import SubdocsEvent, Subscription, TransactionEvent
from ._pycrdt import Transaction as _Transaction
from ._snapshot import Snapshot
from ._transaction import NewTransaction, ReadTransaction, Transaction

T = TypeVar("T", bound=BaseType)
TransactionOrSubdocsEvent = TypeVar(
    "TransactionOrSubdocsEvent", bound=TransactionEvent | SubdocsEvent
)


class Doc(BaseDoc, Generic[T]):
    """
    A shared document.

    All shared types live within the scope of their corresponding documents.
    All updates are generated on a per-document basis.
    All operations on shared types happen in a transaction, whose lifetime is also bound to a
    document.
    """

    def __init__(
        self,
        init: dict[str, T] = {},
        *,
        client_id: int | None = None,
        skip_gc: bool | None = None,
        doc: _Doc | None = None,
        Model=None,
        allow_multithreading: bool = False,
    ) -> None:
        """
        Args:
            init: The initial root types of the document.
            client_id: An optional client ID for the document.
            skip_gc: Whether to skip garbage collection on deleted collections
                on transaction commit.
            allow_multithreading: Whether to allow the document to be used in different threads.
        """
        super().__init__(
            client_id=client_id,
            skip_gc=skip_gc,
            doc=doc,
            Model=Model,
            allow_multithreading=allow_multithreading,
        )
        for k, v in init.items():
            self[k] = v
        if Model is not None:
            self._twin_doc = Doc(init)
        self._send_streams: dict[
            bool, set[MemoryObjectSendStream[TransactionEvent | SubdocsEvent]]
        ] = {False: set(), True: set()}
        self._event_subscription: dict[bool, Subscription] = {}

    @property
    def guid(self) -> int:
        """The GUID of the document."""
        return self._doc.guid()

    @property
    def client_id(self) -> int:
        """The document client ID."""
        return self._doc.client_id()

    def transaction(self, origin: Any = None) -> Transaction:
        """
        Creates a new transaction or gets the current one, if any.
        If an origin is passed and there is already an ongoing transaction,
        the passed origin must be the same as the origin of the current transaction.

        This method must be used with a context manager:

        ```py
        with doc.transaction():
            ...
        ```

        Args:
            origin: An optional origin to set on this transaction.

        Raises:
            RuntimeError: Nested transactions must have same origin as root transaction.

        Returns:
            A new transaction or the current one.
        """
        if self._txn is not None:
            if origin is not None:
                if origin != self._txn.origin:
                    raise RuntimeError(
                        "Nested transactions must have same origin as root transaction"
                    )
            return self._txn
        return Transaction(self, origin=origin)

    def new_transaction(self, origin: Any = None, timeout: float | None = None) -> NewTransaction:
        """
        Creates a new transaction.
        Unlike [transaction()][pycrdt.Doc.transaction], this method will not reuse an ongoing
        transaction.
        If there is already an ongoing transaction, this method will wait (with an optional timeout)
        until the current transaction has finished.
        There are two ways to do so:

        - Use an async context manager:
        ```py
        async with doc.new_transaction():
            ...
        ```
        In this case you most likely access the document in the same thread, which means that
        the [Doc][pycrdt.Doc.__init__] can be created with `allow_multithreading=False`.

        - Use a (sync) context manager:
        ```py
        with doc.new_transaction():
            ...
        ```
        In this case you want to use multithreading, as the ongoing transaction must
        run in another thread (otherwise this will deadlock), which means that
        the [Doc][pycrdt.Doc.__init__] must have been created with `allow_multithreading=True`.

        Args:
            origin: An optional origin to set on this transaction.
            timeout: An optional timeout (in seconds) to acquire a new transaction.

        Raises:
            RuntimeError: Already in a transaction.
            TimeoutError: Could not acquire transaction.

        Returns:
            A new transaction.
        """
        return NewTransaction(self, origin=origin, timeout=timeout)

    def _read_transaction(self, _txn: _Transaction) -> ReadTransaction:
        return ReadTransaction(self, _txn)

    def get_state(self) -> bytes:
        """
        Returns:
            The current document state.
        """
        with self.transaction() as txn:
            assert txn._txn is not None
            return self._doc.get_state(txn._txn)

    def get_update(self, state: bytes | None = None) -> bytes:
        """
        Args:
            state: The optional document state from which to get the update.

        Returns:
            The update from the given document state (if any), or from the document creation.
        """
        if state is None:
            state = b"\x00"
        with self.transaction() as txn:
            assert txn._txn is not None
            return self._doc.get_update(txn._txn, state)

    def apply_update(self, update: bytes) -> None:
        """
        Args:
            update: The update to apply to the document.
        """
        if self._Model is not None:
            twin_doc = cast(Doc, self._twin_doc)
            twin_doc.apply_update(update)
            d = {k: twin_doc[k].to_py() for k in self._Model.model_fields}
            try:
                self._Model(**d)
            except Exception as e:
                self._twin_doc = Doc(dict(self))
                raise e
        with self.transaction() as txn:
            forbid_read_transaction(txn)
            assert txn._txn is not None
            self._doc.apply_update(txn._txn, update)

    @staticmethod
    def from_snapshot(snapshot: "Snapshot", doc: "Doc") -> "Doc":
        """
        Create a new Doc from a Snapshot and an original Doc.
        Args:
            snapshot: The Snapshot to restore to.
            doc: The original Doc to use for options/state.
        Returns:
            A new Doc instance restored to the snapshot state.
        """
        new_doc = _Doc.from_snapshot(snapshot._snapshot, doc._doc)
        return Doc(doc=new_doc)

    def __setitem__(self, key: str, value: T) -> None:
        """
        Sets a document root type:
        ```py
        doc["text"] = Text("Hello")
        ```

        Args:
            key: The name of the root type.
            value: The root type.

        Raises:
            RuntimeError: Key must be of type string.
        """
        if not isinstance(key, str):
            raise RuntimeError("Key must be of type string")
        with self.transaction() as txn:
            forbid_read_transaction(txn)
            integrated = value._get_or_insert(key, self)
            prelim = value._integrate(self, integrated)
            value._init(prelim)

    def __getitem__(self, key: str) -> T:
        """
        Gets the document root type corresponding to the given key:
        ```py
        text = doc["text"]
        ```

        Args:
            key: The key of the root type to get.

        Returns:
            The document root type.
        """
        return self._roots[key]

    def __iter__(self) -> Iterable[str]:
        """
        Returns:
            An iterable over the keys of the document root types.
        """
        return iter(self.keys())

    def get(self, key: str, *, type: type[T]) -> T:
        """
        Gets the document root type corresponding to the given key.
        If it already exists, it will be cast to the given type (if different),
        otherwise a new root type is created.
        ```py
        doc.get("text", type=Text)
        ```

        Returns:
            The root type corresponding to the given key, cast to the given type.
        """
        value = type()
        self[key] = value
        return value

    def keys(self) -> Iterable[str]:
        """
        Returns:
            An iterable over the names of the document root types.
        """
        return self._roots.keys()

    def values(self) -> Iterable[T]:
        """
        Returns:
            An iterable over the document root types.
        """
        return self._roots.values()

    def items(self) -> Iterable[tuple[str, T]]:
        """
        Returns:
            An iterable over the key-value pairs of document root types.
        """
        return self._roots.items()

    @property
    def _roots(self) -> dict[str, T]:
        with self.transaction() as txn:
            assert txn._txn is not None
            return {
                key: (
                    None
                    if val is None
                    else cast(Type[T], base_types[type(val)])(_integrated=val, _doc=self)
                )
                for key, val in self._doc.roots(txn._txn).items()
            }

    def observe(
        self,
        callback: Callable[[TransactionEvent], None]
        | Callable[[TransactionEvent], Awaitable[None]],
    ) -> Subscription:
        """
        Subscribes a callback to be called with the document change event.

        Args:
            callback: The callback to call with the [TransactionEvent][pycrdt.TransactionEvent].
                If the callback is async, async transactions must be used.

        Returns:
            The subscription that can be used to [unobserve()][pycrdt.Doc.unobserve].
        """
        if iscoroutinefunction(callback):
            cb = self._async_callback_to_sync(callback)
        else:
            cb = partial(observe_callback, cast(Callable[[TransactionEvent], None], callback), self)
        subscription = self._doc.observe(cb)
        self._subscriptions.append(subscription)
        return subscription

    def _async_callback_to_sync(
        self,
        async_callback: Callable[[TransactionOrSubdocsEvent], Awaitable[None]],
    ) -> Callable[[TransactionOrSubdocsEvent], None]:
        def callback(event: TransactionOrSubdocsEvent) -> None:
            if self._task_group is None:
                raise RuntimeError("Async callback in non-async transaction")
            self._task_group.start_soon(async_callback, event)

        return callback

    def observe_subdocs(
        self,
        callback: Callable[[SubdocsEvent], None] | Callable[[SubdocsEvent], Awaitable[None]],
    ) -> Subscription:
        """
        Subscribes a callback to be called with the document subdoc change event.

        Args:
            callback: The callback to call with the [SubdocsEvent][pycrdt.SubdocsEvent].

        Returns:
            The subscription that can be used to [unobserve()][pycrdt.Doc.unobserve].
        """
        if iscoroutinefunction(callback):
            cb = self._async_callback_to_sync(callback)
        else:
            cb = partial(observe_callback, cast(Callable[[SubdocsEvent], None], callback), self)
        subscription = self._doc.observe_subdocs(cb)
        self._subscriptions.append(subscription)
        return subscription

    def unobserve(self, subscription: Subscription) -> None:
        """
        Unsubscribes to changes using the given subscription.

        Args:
            subscription: The subscription to unregister.
        """
        self._subscriptions.remove(subscription)
        subscription.drop()

    @overload
    def events(
        self,
        subdocs: Literal[False] = False,
        max_buffer_size: float = float("inf"),
        async_transactions: bool = False,
    ) -> MemoryObjectReceiveStream[TransactionEvent]: ...

    @overload
    def events(
        self,
        subdocs: Literal[True] = True,
        max_buffer_size: float = float("inf"),
        async_transactions: bool = False,
    ) -> MemoryObjectReceiveStream[list[SubdocsEvent]]: ...

    def events(
        self,
        subdocs: bool = False,
        max_buffer_size: float = float("inf"),
        async_transactions: bool = False,
    ):
        """
        Allows to asynchronously iterate over the document events, without using a callback.
        A buffer is used to store the events, allowing to iterate at a (temporarily) slower
        rate than they are produced.

        This method must be used with an async context manager and an async for-loop:

        ```py
        async def main():
            async with doc.events() as events:
                async for event in events:
                    update: bytes = event.update
                    ...
        ```

        Args:
            subdocs: Whether to iterate over the [SubdocsEvent][pycrdt.SubdocsEvent] events
                (default is [TransactionEvent][pycrdt.TransactionEvent]).
            max_buffer_size: Maximum number of events that can be buffered.
            async_transactions: Whether async transactions are used for this document,
                in which case iterating over the events can put back-pressure on the
                transactions (don't use an infinite `max_buffer_size` in this case).

        Returns:
            An async iterator over the document events.
        """
        observe = self.observe_subdocs if subdocs else self.observe
        if not self._send_streams[subdocs]:
            if async_transactions:
                self._event_subscription[subdocs] = observe(
                    partial(self._async_send_event, subdocs)
                )
            else:
                self._event_subscription[subdocs] = observe(partial(self._send_event, subdocs))
        send_stream, receive_stream = create_memory_object_stream[
            Union[TransactionEvent, SubdocsEvent]
        ](max_buffer_size=max_buffer_size)
        self._send_streams[subdocs].add(send_stream)
        return receive_stream

    def _send_event(self, subdocs: bool, event: TransactionEvent | SubdocsEvent):
        to_remove: list[MemoryObjectSendStream[TransactionEvent | SubdocsEvent]] = []
        send_streams = self._send_streams[subdocs]
        for send_stream in send_streams:
            try:
                send_stream.send_nowait(event)
            except BrokenResourceError:
                to_remove.append(send_stream)
        for send_stream in to_remove:
            send_stream.close()
            send_streams.remove(send_stream)
        if not send_streams:
            self.unobserve(self._event_subscription[subdocs])

    async def _async_send_event(self, subdocs: bool, event: TransactionEvent | SubdocsEvent):
        to_remove: list[MemoryObjectSendStream[TransactionEvent | SubdocsEvent]] = []
        send_streams = self._send_streams[subdocs]
        for send_stream in send_streams:
            try:
                await send_stream.send(event)
            except BrokenResourceError:
                to_remove.append(send_stream)
        for send_stream in to_remove:
            send_stream.close()
            send_streams.remove(send_stream)
        if not send_streams:
            self.unobserve(self._event_subscription[subdocs])


class TypedDoc(Typed):
    """
    A container for a [Doc][pycrdt.Doc.__init__] where root shared values have types associated
    with specific keys. The underlying `Doc` can be accessed with the special `_` attribute.

    ```py
    from pycrdt import Array, Doc, Map, Text, TypedDoc

    class MyDoc(TypedDoc):
        map0: Map[int]
        array0: Array[bool]
        text0: Text

    doc = MyDoc()

    doc.map0["foo"] = 3
    doc.array0.append(True)
    doc.text0 += "Hello"
    untyped_doc: Doc = doc._
    ```
    """

    _: Doc

    def __init__(self, doc: TypedDoc | Doc | None = None) -> None:
        super().__init__()
        if doc is None:
            doc = Doc()
        elif isinstance(doc, TypedDoc):
            doc = doc._
        assert isinstance(doc, Doc)
        self._ = doc
        for name, _type in self.__dict__["annotations"].items():
            root_type = _type()
            if isinstance(root_type, Typed):
                root_type = root_type._
            doc[name] = root_type


def observe_callback(
    callback: Callable[[TransactionEvent], None] | Callable[[SubdocsEvent], None],
    doc: Doc,
    event: Any,
) -> None:
    try:
        callback(event)
    except Exception as exc:
        doc._exceptions.append(exc)


base_types[_Doc] = Doc



from __future__ import annotations

import sys
from functools import partial
from types import TracebackType
from typing import TYPE_CHECKING, Any

from anyio import create_task_group, to_thread

from ._pycrdt import Transaction as _Transaction

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: nocover
    from exceptiongroup import ExceptionGroup
    from typing_extensions import Self

if TYPE_CHECKING:
    from ._doc import Doc


class Transaction:
    """
    A read-write transaction that can be used to mutate a document.
    It must be used with a context manager (see [Doc.transaction()][pycrdt.Doc.transaction]):
    ```py
    with doc.transaction():
        ...
    ```
    """

    _doc: Doc
    _txn: _Transaction | None
    _leases: int
    _origin_hash: int | None
    _timeout: float

    def __init__(
        self,
        doc: Doc,
        _txn: _Transaction | None = None,
        *,
        origin: Any = None,
        timeout: float | None = None,
    ) -> None:
        self._doc = doc
        self._txn = _txn
        self._leases = 0
        if origin is None:
            self._origin_hash = None
        else:
            self._origin_hash = hash_origin(origin)
            doc._origins[self._origin_hash] = origin
        self._timeout = -1 if timeout is None else timeout

    def __enter__(self, _acquire_transaction: bool = True) -> Self:
        self._leases += 1
        if self._txn is None:
            if (
                self._doc._allow_multithreading
                and _acquire_transaction
                and not self._doc._txn_lock.acquire(timeout=self._timeout)
            ):
                raise TimeoutError("Could not acquire transaction")
            if self._origin_hash is not None:
                self._txn = self._doc._doc.create_transaction_with_origin(self._origin_hash)
            else:
                self._txn = self._doc._doc.create_transaction()
        self._doc._txn = self
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._leases -= 1
        # only drop the transaction when exiting root context manager
        # since nested transactions reuse the root transaction
        if self._leases == 0:
            assert self._txn is not None
            try:
                if not isinstance(self, ReadTransaction):
                    self._txn.commit()
                    origin_hash = self._txn.origin()
                    if origin_hash is not None:
                        del self._doc._origins[origin_hash]
                    if self._doc._allow_multithreading:
                        self._doc._txn_lock.release()
                    if self._doc._exceptions:
                        exceptions = tuple(self._doc._exceptions)
                        self._doc._exceptions.clear()
                        raise ExceptionGroup("Observer callback error", exceptions)
            finally:
                self._txn.drop()
                self._txn = None
                self._doc._txn = None

    async def __aenter__(self, _acquire_transaction: bool = True) -> Self:
        if self._leases == 0:
            self._doc._task_group = await create_task_group().__aenter__()
        elif self._doc._task_group is None:
            raise RuntimeError("Already in a non-async transaction")
        return self.__enter__(_acquire_transaction)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        self.__exit__(exc_type, exc_val, exc_tb)
        if self._leases == 0:
            assert self._doc._task_group is not None
            res = await self._doc._task_group.__aexit__(exc_type, exc_val, exc_tb)
            self._doc._task_group = None
            return res
        return None

    @property
    def origin(self) -> Any:
        """
        The origin of the transaction.

        Raises:
            RuntimeError: No current transaction.
        """
        if self._txn is None:
            raise RuntimeError("No current transaction")

        origin_hash = self._txn.origin()
        if origin_hash is None:
            return None

        return self._doc._origins[origin_hash]


class NewTransaction(Transaction):
    """
    A read-write transaction that can be used to mutate a document.
    It can be used with a context manager or an async context manager
    (see [Doc.new_transaction()][pycrdt.Doc.new_transaction]):
    ```py
    with doc.new_transaction():
        ...

    async with doc.new_transaction():
        ...
    ```
    """

    async def __aenter__(self) -> Self:  # type: ignore[override]
        if self._doc._allow_multithreading:
            if not await to_thread.run_sync(
                partial(self._doc._txn_lock.acquire, timeout=self._timeout), abandon_on_cancel=True
            ):
                raise TimeoutError("Could not acquire transaction")
        else:
            await self._doc._txn_async_lock.acquire()
        return await super().__aenter__(_acquire_transaction=False)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        res = await super().__aexit__(exc_type, exc_val, exc_tb)
        if not self._doc._allow_multithreading:
            self._doc._txn_async_lock.release()
        return res


class ReadTransaction(Transaction):
    """
    A read-only transaction that cannot be used to mutate a document.
    """


def hash_origin(origin: Any) -> int:
    try:
        return hash(origin)
    except Exception:
        raise TypeError("Origin must be hashable")



from typing import Any, Callable, Generic, Iterator, TypeVar

class Snapshot:
    """A snapshot of a document's state at a given point in time."""

    @staticmethod
    def from_doc(doc: "Doc") -> "Snapshot":
        """Create a snapshot from a document."""

    @staticmethod
    def decode(data: bytes) -> "Snapshot":
        """Decode a snapshot from its binary representation."""

    def encode(self) -> bytes:
        """Encode the snapshot to its binary representation."""

class Doc:
    """Shared document."""

    def __init__(self, client_id: int | None, skip_gc: bool | None) -> None:
        """Create a new document with an optional global client ID.
        If no client ID is passed, a random one will be generated."""

    @staticmethod
    def from_snapshot(snapshot: "Snapshot", doc: Doc) -> "Doc":
        """Create a new Doc from a Snapshot and an original Doc."""

    def client_id(self) -> int:
        """Returns the document unique client identifier."""

    def guid(self) -> int:
        """Returns the document globally unique identifier."""

    def create_transaction(self) -> Transaction:
        """Create a document transaction."""

    def create_transaction_with_origin(self, origin: Any) -> Transaction:
        """Create a document transaction with an origin."""

    def get_or_insert_text(self, txn: Transaction, name: str) -> Text:
        """Create a text root type on this document, or get an existing one."""

    def get_or_insert_array(self, txn: Transaction, name: str) -> Array:
        """Create an array root type on this document, or get an existing one."""

    def get_or_insert_map(self, txn: Transaction, name: str) -> Map:
        """Create a map root type on this document, or get an existing one."""

    def get_or_insert_xml_fragment(self, txn: Transaction, name: str) -> XmlFragment:
        """Create an XML fragment root type on this document, or get an existing one."""

    def get_state(self, txn: Transaction) -> bytes:
        """Get the current document state."""

    def get_update(self, txn: Transaction, state: bytes) -> bytes:
        """Get the update from the given state to the current state."""

    def apply_update(self, txn: Transaction, update: bytes) -> None:
        """Apply the update to the document."""

    def roots(self, txn: Transaction) -> dict[str, Text | Array | Map]:
        """Get top-level (root) shared types available in current document."""

    def observe(self, callback: Callable[[TransactionEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the shared document change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_subdocs(self, callback: Callable[[SubdocsEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the shared document subdoc change event.
        Returns a subscription that can be used to unsubscribe."""

class Subscription:
    """Observer subscription."""

    def drop(self) -> None:
        """Drops the subscription, effectively unobserving."""

class Transaction:
    """Document transaction"""

    def drop(self) -> None:
        """Drop the transaction, effectively committing document changes."""

    def commit(self) -> None:
        """Commit the document changes."""

    @property
    def origin(self) -> Any:
        """The origin of the transaction."""

class TransactionEvent:
    """
    Event generated by the [observe][pycrdt.Doc.observe] method,
    emitted during the transaction commit phase.
    """

    @property
    def update(self) -> bytes:
        """The emitted binary update."""

class SubdocsEvent:
    """
    Event generated by the [observe_subdocs][pycrdt.Doc.observe_subdocs] method,
    emitted during the transaction commit phase."""

class TextEvent:
    """Event generated by `Text.observe` method. Emitted during transaction commit
    phase."""

class ArrayEvent:
    """Event generated by `Array.observe` method. Emitted during transaction commit
    phase."""

class MapEvent:
    """Event generated by `Map.observe` method. Emitted during transaction commit
    phase."""

class XmlEvent:
    """Event generated by `Xml*.observe` methods. Emitted during transaction commit
    phase."""

class Text:
    """Shared text."""

    def len(self, txn: Transaction) -> int:
        """Returns the number of characters visible in the current shared text."""

    def insert(
        self,
        txn: Transaction,
        index: int,
        chunk: str,
        attrs: Iterator[tuple[str, Any]] | None = None,
    ) -> None:
        """Inserts a `chunk` of text at a given `index`, with optional attributes."""

    def insert_embed(
        self,
        txn: Transaction,
        index: int,
        embed: Any,
        attrs: Iterator[tuple[str, Any]] | None = None,
    ) -> None:
        """Inserts an embed at a given `index`, with optional attributes."""

    def format(
        self, txn: Transaction, index: int, len: int, attrs: Iterator[tuple[str, Any]]
    ) -> None:
        """Formats a range of elements."""

    def remove_range(self, txn: Transaction, index: int, len: int) -> None:
        """Removes up to `len` characters from th current shared text, starting at
        given`index`."""

    def get_string(self, txn: Transaction) -> str:
        """Returns a text representation of the current shared text."""

    def diff(self, txn: Transaction) -> list[tuple[Any, dict[str, Any] | None]]:
        """Returns a sequence of formatted chunks."""

    def observe(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the shared text change event.
        Returns a subscription that can be used to unsubscribe."""

    def unobserve(self, subscription: Subscription) -> None:
        """Unsubscribes previously subscribed event callback identified by given
        `subscription`."""

class Array:
    """Shared array."""

    def len(self, txn: Transaction) -> int:
        """Returns the number of elements in the current array."""

    def insert(self, txn: Transaction, index: int, value: Any) -> None:
        """Inserts `value` at the given `index`."""

    def move_to(self, txn: Transaction, source: int, target: int) -> None:
        """Moves element found at `source` index into `target` index position.."""

    def remove_range(self, txn: Transaction, index: int, len: int) -> None:
        """Removes 'len' elements starting at provided `index`."""

    def get(self, txn: Transaction, index: int) -> Any:
        """Retrieves a value stored at a given `index`."""

    def to_json(self, txn: Transaction) -> str:
        """Returns a JSON representation of the current array."""

    def observe(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the array change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_deep(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the array change event
        and its nested elements.
        Returns a subscription that can be used to unsubscribe."""

    def unobserve(self, subscription: Subscription) -> None:
        """Unsubscribes previously subscribed event callback identified by given
        `subscription`."""

class Map:
    """Shared map."""

    def len(self, txn: Transaction) -> int:
        """Returns a number of characters visible in a current text data structure."""

    def insert(self, txn: Transaction, key: str, value: Any) -> None:
        """Inserts `value` at the given `key`."""

    def remove(self, txn: Transaction, key: str) -> None:
        """Removes the `key` entry."""

    def get(self, txn: Transaction, key: str) -> Any:
        """Retrieves a value stored under a given `key`."""

    def to_json(self, txn: Transaction) -> str:
        """Returns a JSON representation of the current map."""

    def observe(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the map change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_deep(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the map change event
        and its nested elements.
        Returns a subscription that can be used to unsubscribe."""

    def unobserve(self, subscription: Subscription) -> None:
        """Unsubscribes previously subscribed event callback identified by given
        `subscription`."""

    def has(self, txn: Transaction, key: str) -> bool:
        """Returns true if the given key exists in the map."""

class XmlFragment:
    def parent(self) -> XmlFragment | XmlElement | XmlText | None:
        """Returns the parent of the XML fragment, if any."""

    def get_string(self, txn: Transaction) -> str:
        """Returns a text representation of the current shared XML."""

    def len(self, txn: Transaction) -> int:
        """Returns the numer of children of the current shared XML."""

    def get(self, txn: Transaction, index: int) -> XmlFragment | XmlElement | XmlText | None:
        """Gets a child item by index, or `None` if the index is out of bounds."""

    def remove_range(self, txn: Transaction, index: int, len: int) -> None:
        """Removes a range of children."""

    def insert_str(self, txn: Transaction, index: int, text: str) -> XmlText:
        """Inserts a text node."""

    def insert_element_prelim(self, txn: Transaction, index: int, tag: str) -> XmlElement:
        """Inserts an empty element node at given index, and returns it."""

    def observe(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_deep(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event
        and its nested elements.
        Returns a subscription that can be used to unsubscribe."""

class XmlElement:
    def parent(self) -> XmlFragment | XmlElement | XmlText | None:
        """Returns the parent of the XML element, if any."""

    def get_string(self, txn: Transaction) -> str:
        """Returns a text representation of the current shared XML."""

    def len(self, txn: Transaction) -> int:
        """Returns the numer of children of the current shared XML."""

    def get(self, txn: Transaction, index: int) -> XmlFragment | XmlElement | XmlText | None:
        """Gets a child item by index, or `None` if the index is out of bounds."""

    def remove_range(self, txn: Transaction, index: int, len: int) -> None:
        """Removes a range of children."""

    def insert_str(self, txn: Transaction, index: int, text: str) -> XmlText:
        """Inserts a text node."""

    def insert_element_prelim(self, txn: Transaction, index: int, tag: str) -> XmlElement:
        """Inserts an empty element node at a given index, and returns it."""

    def attributes(self, txn: Transaction) -> list[tuple[str, str]]:
        """Gets all attributes, as a list of `(key, value)` tuples."""

    def attribute(self, txn: Transaction, name: str) -> str | None:
        """Gets an attribute, or `None` if the attribute does not exist."""

    def insert_attribute(self, txn: Transaction, name: str, value: str) -> None:
        """Inserts or overwrites an attribute."""

    def remove_attribute(self, txn: Transaction, name: str) -> None:
        """Removes an attribute."""

    def siblings(self, txn) -> list[XmlFragment | XmlElement | XmlText]:
        """Gets the siblings of this node."""

    def observe(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_deep(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event
        and its nested elements.
        Returns a subscription that can be used to unsubscribe."""

class XmlText:
    def parent(self) -> XmlFragment | XmlElement | XmlText | None:
        """Returns the XML text parent, if any."""

    def get_string(self, txn: Transaction) -> str:
        """Returns a text representation of the current shared XML."""

    def attributes(self, txn: Transaction) -> list[tuple[str, str]]:
        """Gets all attributes, as a list of `(key, value)` tuples."""

    def attribute(self, txn: Transaction, name: str) -> str | None:
        """Gets an attribute, or `None` if the attribute does not exist."""

    def insert_attribute(self, txn: Transaction, name: str, value: str) -> None:
        """Inserts or overwrites an attribute."""

    def remove_attribute(self, txn: Transaction, name: str) -> None:
        """Removes an attribute."""

    def siblings(self, txn: Transaction) -> list[XmlFragment | XmlElement | XmlText]:
        """Gets the siblings of this node."""

    def insert(
        self,
        txn: Transaction,
        index: int,
        text: str,
        attrs: Iterator[tuple[str, Any]] | None = None,
    ):
        """Inserts text, with optional with attributes."""

    def remove_range(self, txn: Transaction, index: int, len: int):
        """Removes text."""

    def format(self, txn: Transaction, index: int, len: int, attrs: Iterator[tuple[str, Any]]):
        """Adds attributes to a section of text."""

    def diff(self, txn: Transaction) -> list[tuple[Any, dict[str, Any] | None]]:
        """Returns a sequence of formatted chunks."""

    def observe(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event.
        Returns a subscription that can be used to unsubscribe."""

    def observe_deep(self, callback: Callable[[XmlEvent], None]) -> Subscription:
        """Subscribes a callback to be called with the XML change event
        and its nested elements.
        Returns a subscription that can be used to unsubscribe."""

class UndoManager:
    """Undo manager."""

    def __init__(
        self,
        doc: Doc,
        capture_timeout_millis: int,
        timestamp: Callable[[], int],
        undo_stack: list[StackItem] | None = None,
        redo_stack: list[StackItem] | None = None,
    ) -> None:
        """Creates an undo manager."""

    def expand_scope(self, scope: Text | Array | Map) -> None:
        """Extends a list of shared types tracked by current undo manager by a given scope."""

    def include_origin(self, origin: int) -> None:
        """Extends a list of origins tracked by current undo manager by a given origin."""

    def exclude_origin(self, origin: int) -> None:
        """Removes an origin from the list of origins tracked by current undo manager."""

    def can_undo(self) -> bool:
        """Whether there is any change to undo."""

    def undo(self) -> bool:
        """Undo last action tracked by current undo manager."""

    def can_redo(self) -> bool:
        """Whether there is any change to redo."""

    def redo(self) -> bool:
        """Redo last action previously undone by current undo manager."""

    def clear(self) -> None:
        """Clear all items stored within current undo manager."""

    def undo_stack(self) -> list[StackItem]:
        """Returns the undo manager's undo stack."""

    def redo_stack(self) -> list[StackItem]:
        """Returns the undo manager's redo stack."""

class IdSet:
    """A set of deletions in a CRDT document."""

    def __init__(self) -> None:
        """Create a new empty IdSet."""

    def encode(self) -> bytes:
        """Encode the IdSet to bytes."""

    @staticmethod
    def decode(data: bytes) -> IdSet:
        """Decode a IdSet from bytes."""

MetaT = TypeVar("MetaT")

class StackItem(Generic[MetaT]):
    """A unit of work for the [UndoManager][pycrdt.UndoManager], consisting of
    compressed information about all updates and deletions tracked by it.
    """

    def __init__(self, deletions: IdSet, insertions: IdSet, meta: MetaT | None = None) -> None:
        """Create a new StackItem.

        Args:
            deletions: The IdSet of deletions.
            insertions: The IdSet of insertions.
            meta: Optional metadata (can be any Python object).
        """

    @property
    def deletions(self) -> IdSet:
        """Get the deletions IdSet."""

    @property
    def insertions(self) -> IdSet:
        """Get the insertions IdSet."""

    @property
    def meta(self) -> MetaT | None:
        """Custom metadata. Can be any Python object."""

    @staticmethod
    def merge(
        a: "StackItem[MetaT]",
        b: "StackItem[MetaT]",
        merge_meta: Callable[[MetaT | None, MetaT | None], Any] | None = None,
    ) -> "StackItem[Any]":
        """Merge two StackItems into one containing the union of their deletions and insertions.

        Args:
            a: First StackItem to merge.
            b: Second StackItem to merge.
            merge_meta: Optional function to handle metadata conflicts.
                Takes (meta_a, meta_b) and returns the merged metadata.
                If None, keeps the first item's metadata.
        """

class StickyIndex:
    def get_offset(self, txn: Transaction) -> int: ...
    def encode(self) -> bytes: ...
    def to_json_string(self) -> str: ...
    def get_assoc(self) -> int: ...

def merge_updates(updates: tuple[bytes, ...]) -> bytes: ...
def get_state(update: bytes) -> bytes: ...
def get_update(update: bytes, state: bytes) -> bytes: ...
def decode_sticky_index(data: bytes) -> StickyIndex: ...
def get_sticky_index_from_json_string(data: str) -> StickyIndex: ...



from __future__ import annotations

from enum import IntEnum
from typing import Iterator

from ._doc import Doc


class YMessageType(IntEnum):
    """
    A generic Y message type.

    Attributes:
        SYNC: A message type used for synchronizing documents.
        AWARENESS: A message type used for the awareness protocol.
    """

    SYNC = 0
    AWARENESS = 1


class YSyncMessageType(IntEnum):
    """
    A message type used for synchronizing documents.

    Attributes:
        SYNC_STEP1: A synchronization message type used to send a document state.
        SYNC_STEP2: A synchronization message type used to reply to a
            [SYNC_STEP1][pycrdt.YSyncMessageType], consisting of all missing updates and all
            deletions.
        SYNC_UPDATE: A synchronization message type used to send document updates.
    """

    SYNC_STEP1 = 0
    SYNC_STEP2 = 1
    SYNC_UPDATE = 2


def write_var_uint(num: int) -> bytes:
    """
    Encodes a payload length.

    Args:
        num: The payload length to encode.

    Returns:
        The encoded payload length.
    """
    res = []
    while num > 127:
        res.append(128 | (127 & num))
        num >>= 7
    res.append(num)
    return bytes(res)


def create_awareness_message(data: bytes) -> bytes:
    """
    Creates an [AWARENESS][pycrdt.YMessageType] message.

    Args:
        data: The data to send in the message.

    Returns:
        The [AWARENESS][pycrdt.YMessageType] message.
    """
    return bytes([YMessageType.AWARENESS]) + write_message(data)


def create_message(data: bytes, msg_type: int) -> bytes:
    """
    Creates a SYNC message.

    Args:
        data: The data to send in the message.
        msg_type: The [SYNC message type][pycrdt.YSyncMessageType].

    Returns:
        The SYNC message.
    """
    return bytes([YMessageType.SYNC, msg_type]) + write_message(data)


def create_sync_step1_message(data: bytes) -> bytes:
    """
    Creates a [SYNC_STEP1][pycrdt.YSyncMessageType.SYNC_STEP1] message consisting
    of a document state.

    Args:
        data: The document state.

    Returns:
        A [SYNC_STEP1][pycrdt.YSyncMessageType.SYNC_STEP1] message.
    """
    return create_message(data, YSyncMessageType.SYNC_STEP1)


def create_sync_step2_message(data: bytes) -> bytes:
    """
    Creates a [SYNC_STEP2][pycrdt.YSyncMessageType.SYNC_STEP2] message in
    reply to a [SYNC_STEP1][pycrdt.YSyncMessageType.SYNC_STEP1] message.

    Args:
        data: All missing updates and deletetions.

    Returns:
        A [SYNC_STEP2][pycrdt.YSyncMessageType.SYNC_STEP2] message.
    """
    return create_message(data, YSyncMessageType.SYNC_STEP2)


def create_update_message(data: bytes) -> bytes:
    """
    Creates a [SYNC_UPDATE][pycrdt.YSyncMessageType] message that
    contains a document update.

    Args:
        data: The document update.

    Returns:
        A [SYNC_UPDATE][pycrdt.YSyncMessageType] message.
    """
    return create_message(data, YSyncMessageType.SYNC_UPDATE)


class Encoder:
    """
    An encoder capable of writing messages to a binary stream.
    """

    stream: list[bytes]

    def __init__(self) -> None:
        self.stream = []

    def write_var_uint(self, num: int) -> None:
        """
        Encodes a number.

        Args:
            num: The number to encode.
        """
        self.stream.append(write_var_uint(num))

    def write_var_string(self, text: str) -> None:
        """
        Encodes a string.

        Args:
            text: The string to encode.
        """
        self.stream.append(write_var_uint(len(text)))
        self.stream.append(text.encode())

    def to_bytes(self) -> bytes:
        """
        Returns:
            The binary stream.
        """
        return b"".join(self.stream)


class Decoder:
    """
    A decoder capable of reading messages from a byte stream.
    """

    def __init__(self, stream: bytes):
        """
        Args:
            stream: The byte stream from which to read messages.
        """
        self.stream = stream
        self.length = len(stream)
        self.i0 = 0

    def read_var_uint(self) -> int:
        """
        Decodes the current message length.

        Returns:
            The decoded length of the message.
        """
        if self.length <= 0:
            raise RuntimeError("Y protocol error")
        uint = 0
        i = 0
        while True:
            byte = self.stream[self.i0]
            uint += (byte & 127) << i
            i += 7
            self.i0 += 1
            self.length -= 1
            if byte < 128:
                break
        return uint

    def read_message(self) -> bytes | None:
        """
        Reads a message from the byte stream, ready to read the next message if any.

        Returns:
            The current message, if any.
        """
        if self.length == 0:
            return None
        length = self.read_var_uint()
        if length == 0:
            return b""
        i1 = self.i0 + length
        message = self.stream[self.i0 : i1]
        self.i0 = i1
        self.length -= length
        return message

    def read_messages(self) -> Iterator[bytes]:
        """
        A generator that reads messages from the byte stream.

        Returns:
            A generator that yields messages.
        """
        while True:
            message = self.read_message()
            if message is None:
                return
            yield message

    def read_var_string(self) -> str:
        """
        Reads a message as an UTF-8 string from the byte stream, ready to read the next message if
        any.

        Returns:
            The current message as a string.
        """
        message = self.read_message()
        if message is None:
            return ""
        return message.decode("utf-8")


def read_message(stream: bytes) -> bytes:
    """
    Reads a message from a byte stream.

    Args:
        stream: The byte stream from which to read the message.

    Returns:
        The message read from the byte stream.
    """
    message = Decoder(stream).read_message()
    assert message is not None
    return message


def write_message(stream: bytes) -> bytes:
    """
    Writes a stream in a message.

    Args:
        stream: The byte stream to write in a message.

    Returns:
        The message containing the stream.
    """
    return write_var_uint(len(stream)) + stream


def handle_sync_message(message: bytes, ydoc: Doc) -> bytes | None:
    """
    Processes a [synchronization message][pycrdt.YSyncMessageType] on a document.

    Args:
        message: A synchronization message.
        ydoc: The [Doc][pycrdt.Doc] that this message targets.

    Returns:
        The [SYNC_STEP2][pycrdt.YSyncMessageType] reply message, if the message
        was a [SYNC_STEP1][pycrdt.YSyncMessageType].
    """
    message_type = message[0]
    msg = message[1:]

    if message_type == YSyncMessageType.SYNC_STEP1:
        state = read_message(msg)
        update = ydoc.get_update(state)
        reply = create_sync_step2_message(update)
        return reply

    if message_type in (
        YSyncMessageType.SYNC_STEP2,
        YSyncMessageType.SYNC_UPDATE,
    ):
        update = read_message(msg)
        # Ignore empty updates
        if update != b"\x00\x00":
            ydoc.apply_update(update)

    return None


def create_sync_message(ydoc: Doc) -> bytes:
    """
    Creates a [SYNC_STEP1][pycrdt.YSyncMessageType] message that
    contains the state of a [Doc][pycrdt.Doc].

    Args:
        ydoc: The [Doc][pycrdt.Doc] for which to create the message.

    Returns:
        A [SYNC_STEP1][pycrdt.YSyncMessageType] message.
    """
    state = ydoc.get_state()
    message = create_sync_step1_message(state)
    return message



from __future__ import annotations

import json
import sys
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

from ._pycrdt import StickyIndex as _StickyIndex
from ._pycrdt import decode_sticky_index, get_sticky_index_from_json_string

if TYPE_CHECKING:
    from pycrdt import Transaction

    from ._base import Sequence


class Assoc(IntEnum):
    """
    Whether to associate a sticky index with the item on its left (`BEFORE`)
    or on its right (`AFTER`).
    """

    AFTER = 0
    BEFORE = -1


@dataclass
class StickyIndex:
    """
    A permanent position that sticks to the same place even when concurrent updates are made.
    """

    _sticky_index: _StickyIndex
    _sequence: Sequence | None = None

    def get_index(self, transaction: Transaction | None = None) -> int:
        """
        Get the current value of the index.

        Args:
            transaction: A transaction that must be provided when the sticky index
                was deserialized and not associated with a shared type.

        Returns:
            The current index.

        Raises:
            RuntimeError: No transaction was provided and no shared type was associated
                with the deserialized sticky index.
        """
        if transaction is not None:
            _txn = transaction._txn
            assert _txn is not None
            return self._sticky_index.get_offset(_txn)

        if self._sequence is not None:
            with self._sequence.doc.transaction() as txn:
                _txn = txn._txn
                assert _txn is not None
                return self._sticky_index.get_offset(_txn)

        raise RuntimeError("No transaction available")

    @property
    def assoc(self) -> Assoc:
        """
        The [Assoc][pycrdt.Assoc] of the sticky index (before or after).
        """
        return Assoc(self._sticky_index.get_assoc())

    def encode(self) -> bytes:
        """
        Encode the sticky index to binary.

        Returns:
            The binary representation of the sticky index.
        """
        return self._sticky_index.encode()

    def to_json(self) -> dict:
        """
        Serialize the sticky index to JSON.

        Returns:
            The JSON representation of the sticky index.
        """
        return json.loads(self._sticky_index.to_json_string())

    @classmethod
    def new(cls, sequence: Sequence, index: int, assoc: Assoc) -> Self:
        """
        Create a sticky index before or after the specified index.

        Args:
            sequence: The [Array][pycrdt.Array] or [Text][pycrdt.Text] to get the sticky index from.
            index: The index at which the sticky index should remain.
            assoc: The [Assoc][pycrdt.Assoc] the sticky index should be associated
                with (before or after).

        Returns:
            The sticky index.
        """
        with sequence.doc.transaction() as txn:
            self = cls(sequence.integrated.sticky_index(txn._txn, index, assoc), sequence)
            return self

    @classmethod
    def decode(cls, data: bytes, sequence: Sequence | None = None) -> Self:
        """
        Create the sticky index from its binary representation.

        Args:
            data: The binary data to get the sticky index from.
            sequence: The [Array][pycrdt.Array] or [Text][pycrdt.Text] the sticky index belongs to.
                If not provided, a [Transaction][pycrdt.Transaction] will be needed when getting
                the index.

        Returns:
            The decoded sticky index.
        """
        self = cls(decode_sticky_index(data), sequence)
        return self

    @classmethod
    def from_json(cls, data: dict, sequence: Sequence | None = None) -> Self:
        """
        Create a sticky index from its JSON representation.

        Args:
            data: The JSON dictionary to get the sticky index from.
            sequence: The [Array][pycrdt.Array] or [Text][pycrdt.Text] the sticky index belongs to.
                If not provided, a [Transaction][pycrdt.Transaction] will be needed when getting
                the index.

        Returns:
            The deserialized sticky index.
        """
        self = cls(get_sticky_index_from_json_string(json.dumps(data)), sequence)
        return self



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, cast, overload

from ._base import BaseDoc, BaseEvent, BaseType, Sequence, Typed, base_types, event_types
from ._pycrdt import Array as _Array
from ._pycrdt import ArrayEvent as _ArrayEvent
from ._pycrdt import Subscription

if TYPE_CHECKING:
    from ._doc import Doc

T = TypeVar("T")


class Array(Sequence, Generic[T]):
    """
    A collection used to store data in an indexed sequence structure, similar to a Python `list`.
    """

    _prelim: list[T] | None
    _integrated: _Array | None

    def __init__(
        self,
        init: list[T] | None = None,
        *,
        _doc: Doc | None = None,
        _integrated: _Array | None = None,
    ) -> None:
        """
        Creates an array with an optional initial value:
        ```py
        array0 = Array()
        array1 = Array(["foo", 3, array0])
        ```

        Args:
            init: The list from which to initialize the array.
        """
        super().__init__(
            init=init,
            _doc=_doc,
            _integrated=_integrated,
        )

    def _init(self, value: list[T] | None) -> None:
        if value is None:
            return
        with self.doc.transaction():
            for i, v in enumerate(value):
                self._set(i, v)

    def _set(self, index: int, value: T) -> None:
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(value, BaseDoc):
                # subdoc
                self.integrated.insert_doc(txn._txn, index, value._doc)
            elif isinstance(value, BaseType):
                # shared type
                assert txn._txn is not None
                self._do_and_integrate("insert", value, txn._txn, index)
            else:
                # primitive type
                self.integrated.insert(txn._txn, index, value)

    def _get_or_insert(self, name: str, doc: Doc) -> _Array:
        assert doc._txn is not None
        assert doc._txn._txn is not None
        return doc._doc.get_or_insert_array(doc._txn._txn, name)

    def __len__(self) -> int:
        """
        ```py
        Doc()["array"] = array = Array([2, 3, 0])
        assert len(array) == 3
        ```

        Returns:
            The length of the array.
        """
        with self.doc.transaction() as txn:
            return self.integrated.len(txn._txn)

    def append(self, value: T) -> None:
        """
        Appends an item to the array.

        Args:
            value: The item to append to the array.
        """
        with self.doc.transaction():
            self += [value]

    def extend(self, value: list[T]) -> None:
        """
        Extends the array with a list of items.

        Args:
            value: The items that will extend the array.
        """
        with self.doc.transaction():
            self += value

    def clear(self) -> None:
        """
        Removes all items from the array.
        """
        del self[:]

    def insert(self, index: int, object: T) -> None:
        """
        Inserts an item at a given index in the array.

        Args:
            index: The index where to insert the item.
            object: The item to insert in the array.
        """
        self[index:index] = [object]

    def pop(self, index: int = -1) -> T:
        """
        Removes the item at the given index from the array, and returns it.
        If no index is passed, removes and returns the last item.

        Args:
            index: The optional index of the item to pop.

        Returns:
            The item at the given index, or the last item.
        """
        with self.doc.transaction():
            index = self._check_index(index)
            res = self[index]
            if isinstance(res, BaseType):
                res = res.to_py()
            del self[index]
            return res

    def move(self, source_index: int, destination_index: int) -> None:
        """
        Moves an item in the array from a source index to a destination index.

        Args:
            source_index: The index of the item to move.
            destination_index: The index where the item will be inserted.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            source_index = self._check_index(source_index)
            destination_index = self._check_index(destination_index)
            self.integrated.move_to(txn._txn, source_index, destination_index)

    def __add__(self, value: list[T]) -> Array[T]:
        """
        Extends the array with a list of items:
        ```py
        Doc()["array"] = array = Array(["foo"])
        array += ["bar", "baz"]
        assert array.to_py() == ["foo", "bar", "baz"]
        ```

        Args:
            value: The items that will extend the array.

        Returns:
            The extended array.
        """
        with self.doc.transaction():
            length = len(self)
            self[length:length] = value
            return self

    def __radd__(self, value: list[T]) -> Array[T]:
        """
        Prepends a list of items to the array:
        ```py
        Doc()["array"] = array = Array(["bar", "baz"])
        array = ["foo"] + array
        assert array.to_py() == ["foo", "bar", "baz"]
        ```

        Args:
            value: The list of items to prepend.

        Returns:
            The prepended array.
        """
        with self.doc.transaction():
            self[0:0] = value
            return self

    @overload
    def __setitem__(self, key: int, value: T) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: list[T]) -> None: ...

    def __setitem__(self, key, value):
        """
        Replaces the item at the given index with a new item:
        ```py
        Doc()["array"] = array = Array(["foo", "bar"])
        array[1] = "baz"
        assert array.to_py() == ["foo", "baz"]
        ```

        Args:
            key: The index of the item to replace.
            value: The new item to set.

        Raises:
            RuntimeError: Index must be of type integer.
        """
        with self.doc.transaction():
            if isinstance(key, int):
                key = self._check_index(key)
                del self[key]
                self[key:key] = [value]
            elif isinstance(key, slice):
                if key.step is not None:
                    raise RuntimeError("Step not supported")
                if key.start != key.stop:
                    raise RuntimeError("Start and stop must be equal")
                if key.start > len(self) or key.start < 0:
                    raise RuntimeError("Index out of range")
                for i, v in enumerate(value):
                    self._set(i + key.start, v)
            else:
                raise RuntimeError("Index must be of type integer")

    def _check_index(self, idx: int) -> int:
        if not isinstance(idx, int):
            raise RuntimeError("Index must be of type integer")
        length = len(self)
        if idx < 0:
            idx += length
        if idx < 0 or idx >= length:
            raise IndexError("Array index out of range")
        return idx

    def __delitem__(self, key: int | slice) -> None:
        """
        Removes the item at the given index from the array:
        ```py
        Doc()["array"] = array = Array(["foo", "bar", "baz"])
        del array[2]
        assert array.to_py() == ["foo", "bar"]
        ```

        Args:
            key: The index of the item to remove.

        Raises:
            RuntimeError: Array indices must be integers or slices.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(key, int):
                key = self._check_index(key)
                self.integrated.remove_range(txn._txn, key, 1)
            elif isinstance(key, slice):
                if key.step is not None:
                    raise RuntimeError("Step not supported")
                if key.start is None:
                    i = 0
                elif key.start < 0:
                    raise RuntimeError("Negative start not supported")
                else:
                    i = key.start
                if key.stop is None:
                    n = len(self) - i
                elif key.stop < 0:
                    raise RuntimeError("Negative stop not supported")
                else:
                    n = key.stop - i
                self.integrated.remove_range(txn._txn, i, n)
            else:
                raise TypeError(
                    f"Array indices must be integers or slices, not {type(key).__name__}"
                )

    @overload
    def __getitem__(self, key: int) -> T: ...

    @overload
    def __getitem__(self, key: slice) -> list[T]: ...

    def __getitem__(self, key):
        """
        Gets the item at the given index:
        ```py
        Doc()["array"] = array = Array(["foo", "bar", "baz"])
        assert array[1] == "bar"
        ```

        Returns:
            The item at the given index.
        """
        with self.doc.transaction() as txn:
            if isinstance(key, int):
                key = self._check_index(key)
                return self._maybe_as_type_or_doc(self.integrated.get(txn._txn, key))
            elif isinstance(key, slice):
                i0 = 0 if key.start is None else key.start
                i1 = len(self) if key.stop is None else key.stop
                step = 1 if key.step is None else key.step
                return [self[i] for i in range(i0, i1, step)]

    def __iter__(self) -> ArrayIterator:
        """
        ```py
        Doc()["array"] = array = Array(["foo", "foo"])
        for value in array:
            assert value == "foo"
        ```
        Returns:
            An iterable over the items of the array.
        """
        return ArrayIterator(self)

    def __contains__(self, item: T) -> bool:
        """
        Checks if the given item is in the array:
        ```py
        Doc()["array"] = array = Array(["foo", "bar"])
        assert "baz" not in array
        ```

        Args:
            item: The item to look for in the array.

        Returns:
            True if the item was found.
        """
        return item in [value for value in self]

    def __str__(self) -> str:
        """
        ```py
        Doc()["array"] = array = Array([2, 3, 0])
        assert str(array) == "[2,3,0]"
        ```

        Returns:
            The string representation of the array.
        """
        with self.doc.transaction() as txn:
            return self.integrated.to_json(txn._txn)

    def to_py(self) -> list[T] | None:
        """
        Recursively converts the array's items to Python objects, and
        returns them in a list. If the array was not yet inserted in a document,
        returns `None` if the array was not initialized.

        Returns:
            The array recursively converted to Python objects, or `None`.
        """
        if self._integrated is None:
            py = self._prelim
            if py is None:
                return None
        else:
            py = [value for value in self]
        for idx, val in enumerate(py):
            if isinstance(val, BaseType):
                py[idx] = val.to_py()
        return py

    def observe(self, callback: Callable[[ArrayEvent], None]) -> Subscription:
        """
        Subscribes a callback to be called with the array event.

        Args:
            callback: The callback to call with the [ArrayEvent][pycrdt.ArrayEvent].
        """
        return super().observe(cast(Callable[[BaseEvent], None], callback))


class ArrayEvent(BaseEvent):
    """
    An array change event.

    Attributes:
        target (Array): The changed array.
        delta (list[dict[str, Any]]): A list of items describing the changes.
        path (list[int | str]): A list with the indices pointing to the array that was changed.
        transaction (ReadTransaction): The transaction this change was done in.
    """

    __slots__ = "target", "delta", "path", "transaction"


class ArrayIterator:
    def __init__(self, array: Array):
        self.array = array
        self.length = len(array)
        self.idx = 0

    def __iter__(self) -> ArrayIterator:
        return self  # pragma: nocover

    def __next__(self) -> Any:
        if self.idx == self.length:
            raise StopIteration

        res = self.array[self.idx]
        self.idx += 1
        return res


class TypedArray(Typed, Generic[T]):
    """
    A container for an [Array][pycrdt.Array.__init__] where values have types that can be
    other typed containers, e.g. a [TypedMap][pycrdt.TypedMap]. The subclass of `TypedArray[T]`
    must have a special `type: T` annotation where `T` is the same type.
    The underlying `Array` can be accessed with the special `_` attribute.

    ```py
    from pycrdt import Array, TypedArray, TypedDoc, TypedMap

    class MyMap(TypedMap):
        name: str
        toggle: bool
        nested: Array[bool]

    class MyArray(TypedArray[MyMap]):
        type: MyMap

    class MyDoc(TypedDoc):
        array0: MyArray

    doc = MyDoc()

    map0 = MyMap()
    doc.array0.append(map0)
    map0.name = "foo"
    map0.toggle = True
    map0.nested = Array([True, False])

    print(doc.array0._.to_py())
    # [{'name': 'foo', 'toggle': True, 'nested': [True, False]}]
    print(doc.array0[0].name)
    # foo
    print(doc.array0[0].toggle)
    # True
    print(doc.array0[0].nested.to_py())
    # [True, False]
    ```
    """

    type: T
    _: Array

    def __init__(self, array: TypedArray | Array | None = None) -> None:
        super().__init__()
        if array is None:
            array = Array()
        elif isinstance(array, TypedArray):
            array = array._
        self._ = array
        self.__dict__["type"] = self.__dict__["annotations"]["type"]

    def __getitem__(self, key: int) -> T:
        return self.__dict__["type"](self._[key])

    def __setitem__(self, key: int, value: T) -> None:
        item = value._ if isinstance(value, Typed) else value
        self._[key] = item

    def append(self, value: T) -> None:
        item = value._ if isinstance(value, Typed) else value
        self._.append(item)

    def extend(self, value: list[T]) -> None:
        items = [item._ if isinstance(item, Typed) else item for item in value]
        self._.extend(items)

    def __len__(self) -> int:
        return len(self._)


base_types[_Array] = Array
event_types[_ArrayEvent] = ArrayEvent



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from ._array import Array as Array
from ._array import ArrayEvent as ArrayEvent
from ._array import TypedArray as TypedArray
from ._awareness import Awareness as Awareness
from ._awareness import is_awareness_disconnect_message as is_awareness_disconnect_message
from ._doc import Doc as Doc
from ._doc import TypedDoc as TypedDoc
from ._map import Map as Map
from ._map import MapEvent as MapEvent
from ._map import TypedMap as TypedMap
from ._provider import Channel as Channel
from ._provider import Provider as Provider
from ._pycrdt import IdSet as IdSet
from ._pycrdt import StackItem as StackItem
from ._pycrdt import SubdocsEvent as SubdocsEvent
from ._pycrdt import Subscription as Subscription
from ._pycrdt import TransactionEvent as TransactionEvent
from ._snapshot import Snapshot as Snapshot
from ._sticky_index import Assoc as Assoc
from ._sticky_index import StickyIndex as StickyIndex
from ._sync import Decoder as Decoder
from ._sync import Encoder as Encoder
from ._sync import YMessageType as YMessageType
from ._sync import YSyncMessageType as YSyncMessageType
from ._sync import create_awareness_message as create_awareness_message
from ._sync import create_sync_message as create_sync_message
from ._sync import create_update_message as create_update_message
from ._sync import handle_sync_message as handle_sync_message
from ._sync import read_message as read_message
from ._sync import write_message as write_message
from ._sync import write_var_uint as write_var_uint
from ._text import Text as Text
from ._text import TextEvent as TextEvent
from ._transaction import NewTransaction as NewTransaction
from ._transaction import ReadTransaction as ReadTransaction
from ._transaction import Transaction as Transaction
from ._undo import UndoManager as UndoManager
from ._update import get_state as get_state
from ._update import get_update as get_update
from ._update import merge_updates as merge_updates
from ._version import __version__ as __version__
from ._xml import XmlElement as XmlElement
from ._xml import XmlEvent as XmlEvent
from ._xml import XmlFragment as XmlFragment
from ._xml import XmlText as XmlText



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Iterator, cast

from ._base import BaseEvent, Sequence, base_types, event_types
from ._pycrdt import Subscription
from ._pycrdt import Text as _Text
from ._pycrdt import TextEvent as _TextEvent

if TYPE_CHECKING:
    from ._doc import Doc


class Text(Sequence):
    """
    A shared data type used for collaborative text editing, similar to a Python `str`.
    """

    _prelim: str | None
    _integrated: _Text | None

    def __init__(
        self,
        init: str | None = None,
        *,
        _doc: Doc | None = None,
        _integrated: _Text | None = None,
    ) -> None:
        """
        Creates a text with an optional initial value:
        ```py
        text = Text("Hello, World!")
        ```

        Args:
            init: The string from which to initialize the text.
        """
        super().__init__(
            init=init,
            _doc=_doc,
            _integrated=_integrated,
        )

    def _init(self, value: str | None) -> None:
        if value is None:
            return
        with self.doc.transaction() as txn:
            self.integrated.insert(txn._txn, 0, value)

    def _get_or_insert(self, name: str, doc: Doc) -> _Text:
        assert doc._txn is not None
        assert doc._txn._txn is not None
        return doc._doc.get_or_insert_text(doc._txn._txn, name)

    def __iter__(self) -> Iterator[str]:
        """
        ```py
        Doc()["text"] = text = Text("***")
        for character in text:
            assert character == "*"
        ```

        Returns:
            An iterable over the characters of the text.
        """
        return iter(str(self))

    def __contains__(self, item: str) -> bool:
        """
        Checks if the given string is in the text:
        ```py
        Doc()["text"] = text = Text("Hello, World!")
        assert "World" in text
        ```

        Args:
            item: The string to look for in the text.

        Returns:
            True if the string was found.
        """
        return item in str(self)

    def __len__(self) -> int:
        """
        ```py
        Doc()["text"] = text = Text("Hello")
        assert len(text) == 5
        ```

        Returns:
            The length of the text.
        """
        with self.doc.transaction() as txn:
            return self.integrated.len(txn._txn)

    def __str__(self) -> str:
        """
        Returns:
            The text as a Python `str`.
        """
        with self.doc.transaction() as txn:
            return self.integrated.get_string(txn._txn)

    def to_py(self) -> str | None:
        """
        Returns:
            The text as a Python `str`.
        """
        if self._integrated is None:
            return self._prelim
        return str(self)

    def __iadd__(self, value: str) -> Text:
        """
        Concatenates a string to the text:
        ```py
        Doc()["text"] = text = Text("Hello")
        text += ", World!"
        assert str(text) == "Hello, World!"
        ```

        Args:
            value: The string to concatenate.

        Returns:
            The concatenated text.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            self.integrated.insert(txn._txn, len(self), value)
            return self

    def _check_slice(self, key: slice) -> tuple[int, int]:
        if key.step is not None:
            raise RuntimeError("Step not supported")
        if key.start is None:
            start = 0
        elif key.start < 0:
            raise RuntimeError("Negative start not supported")
        else:
            start = key.start
        if key.stop is None:
            stop = len(self)
        elif key.stop < 0:
            raise RuntimeError("Negative stop not supported")
        else:
            stop = key.stop
        return start, stop

    def __delitem__(self, key: int | slice) -> None:
        """
        Removes the characters at the given index or slice:
        ```py
        Doc()["text"] = text = Text("Hello, World!")
        del text[5]
        assert str(text) == "Hello World!"
        del text[5:]
        assert str(text) == "Hello"
        ```

        Args:
            key: The index or the slice of the characters to remove.

        Raises:
            RuntimeError: Step not supported.
            RuntimeError: Negative start not supported.
            RuntimeError: Negative stop not supported.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(key, int):
                self.integrated.remove_range(txn._txn, key, 1)
            elif isinstance(key, slice):
                start, stop = self._check_slice(key)
                length = stop - start
                if length > 0:
                    self.integrated.remove_range(txn._txn, start, length)
            else:
                raise RuntimeError(f"Index not supported: {key}")

    def __getitem__(self, key: int | slice) -> str:
        """
        Gets the characters at the given index or slice:
        ```py
        Doc()["text"] = text = Text("Hello, World!")
        assert text[:5] == "Hello"
        ```

        Returns:
            The characters at the given index or slice.
        """
        value = str(self)
        return value[key]

    def __setitem__(self, key: int | slice, value: str) -> None:
        """
        Replaces the characters at the given index or slice with new characters:
        ```py
        Doc()["text"] = text = Text("Hello, World!")
        text[7:12] = "Brian"
        assert text == "Hello, Brian!"
        ```

        Args:
            key: The index or slice of the characters to replace.
            value: The new characters to set.

        Raises:
            RuntimeError: Step not supported.
            RuntimeError: Negative start not supported.
            RuntimeError: Negative stop not supported.
            RuntimeError: Single item assigned value must have a length of 1.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(key, int):
                value_len = len(value)
                if value_len != 1:
                    raise RuntimeError(
                        f"Single item assigned value must have a length of 1, not {value_len}"
                    )
                del self[key]
                self.integrated.insert(txn._txn, key, value)
            elif isinstance(key, slice):
                start, stop = self._check_slice(key)
                length = stop - start
                if length > 0:
                    self.integrated.remove_range(txn._txn, start, length)
                self.integrated.insert(txn._txn, start, value)
            else:
                raise RuntimeError(f"Index not supported: {key}")

    def clear(self) -> None:
        """Remove the entire range of characters."""
        del self[:]

    def insert(self, index: int, value: str, attrs: dict[str, Any] | None = None) -> None:
        """
        Inserts a string at a given index in the text, with optional attributes.
        ```py
        Doc()["text"] = text = Text("Hello World!")
        text.insert(5, ",")
        assert text == "Hello, World!"
        ```

        Args:
            index: The index where to insert the string.
            value: The string to insert in the text.
            attrs: The formatting of attributes to apply.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            self.integrated.insert(
                txn._txn, index, value, iter(attrs.items()) if attrs is not None else None
            )

    def insert_embed(self, index: int, value: Any, attrs: dict[str, Any] | None = None) -> None:
        """
        Insert a value as an embed at a given index in the text, with optional attributes.

        Args:
            index: The index where to insert the embedded content.
            value: The value to embed.
            attrs: The formatting attributes to apply.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            self.integrated.insert_embed(
                txn._txn, index, value, iter(attrs.items()) if attrs is not None else None
            )

    def format(self, start: int, stop: int, attrs: dict[str, Any]) -> None:
        """
        Applies formatting attributes to a section of text between given start and stop indices.

        Args:
            start: The index at which to start applying attributes (included).
            stop: The index at which to stop applying attributes (excluded).
            attrs: The formatting attributes to apply.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            start, stop = self._check_slice(slice(start, stop))
            length = stop - start
            if length > 0:
                self.integrated.format(txn._txn, start, length, iter(attrs.items()))

    def diff(self) -> list[tuple[Any, dict[str, Any] | None]]:
        """
        Returns:
            A list of formatted chunks that the current text corresponds to.
                Each list item is a tuple containing the chunk's content and formatting attributes.
                The content is usually the text as a string, but may be other data for embedded
                objects.
        """
        with self.doc.transaction() as txn:
            return self.integrated.diff(txn._txn)

    def observe(self, callback: Callable[[TextEvent], None]) -> Subscription:
        """
        Subscribes a callback to be called with the text event.

        Args:
            callback: The callback to call with the [TextEvent][pycrdt.TextEvent].
        """
        return super().observe(cast(Callable[[BaseEvent], None], callback))


class TextEvent(BaseEvent):
    """
    A text change event.

    Attributes:
        target (Text): The changed text.
        delta (list[dict[str, Any]]): A list of items describing the changes.
        path (list[int | str]): A list with the indices pointing to the text that was changed.
        transaction (ReadTransaction): The transaction this change was done in.
    """

    __slots__ = "target", "delta", "path", "transaction"


base_types[_Text] = Text
event_types[_TextEvent] = TextEvent



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, overload

from ._base import BaseEvent, BaseType, base_types, event_types
from ._pycrdt import XmlElement as _XmlElement
from ._pycrdt import XmlEvent as _XmlEvent
from ._pycrdt import XmlFragment as _XmlFragment
from ._pycrdt import XmlText as _XmlText

if TYPE_CHECKING:
    from typing import Any, Iterable, Mapping, Sized, TypeVar

    from ._doc import Doc

    T = TypeVar("T")


def _integrated_to_wrapper(
    doc: Doc, inner: _XmlText | _XmlElement | _XmlFragment
) -> XmlText | XmlElement | XmlFragment:
    if isinstance(inner, _XmlElement):
        return XmlElement(_doc=doc, _integrated=inner)
    if isinstance(inner, _XmlFragment):
        return XmlFragment(_doc=doc, _integrated=inner)
    return XmlText(_doc=doc, _integrated=inner)


def _check_slice(value: Sized, key: slice) -> tuple[int, int]:
    if key.step is not None:
        raise RuntimeError("Step not supported")
    if key.start is None:
        start = 0
    elif key.start < 0:
        raise RuntimeError("Negative start not supported")
    else:
        start = key.start
    if key.stop is None:
        stop = len(value)
    elif key.stop < 0:
        raise RuntimeError("Negative stop not supported")
    else:
        stop = key.stop
    return start, stop


class _XmlBaseMixin(BaseType):
    _integrated: _XmlElement | _XmlText | _XmlFragment | None

    @property
    def parent(self) -> XmlFragment | XmlElement | XmlText | None:
        """
        The parent of this node, if any.
        """
        inner = self.integrated.parent()
        if inner is None:
            return None
        return _integrated_to_wrapper(self.doc, inner)

    def __str__(self):
        with self.doc.transaction() as txn:
            return self.integrated.get_string(txn._txn)

    def __eq__(self, other: object):
        if not isinstance(other, _XmlBaseMixin):
            return False
        return self.integrated == other.integrated

    def __hash__(self) -> int:
        return hash(self.integrated)


class _XmlFragmentTraitMixin(_XmlBaseMixin):
    _integrated: _XmlElement | _XmlFragment | None

    @property
    def children(self) -> XmlChildrenView:
        """
        A list-like view into this object's child nodes.
        """
        return XmlChildrenView(self)


class _XmlTraitMixin(_XmlBaseMixin):
    _integrated: _XmlElement | _XmlText | None

    @property
    def attributes(self) -> XmlAttributesView:
        """
        A dict-like view into this object's attributes.
        """
        return XmlAttributesView(self)


class XmlFragment(_XmlFragmentTraitMixin):
    _prelim: list[XmlFragment | XmlElement | XmlText] | None
    _integrated: _XmlFragment | None

    def __init__(
        self,
        init: Iterable[XmlFragment | XmlElement | XmlText] | None = None,
        *,
        _doc: Doc | None = None,
        _integrated: _XmlFragment | None = None,
    ) -> None:
        super().__init__(
            init=list(init) if init is not None else None,
            _doc=_doc,
            _integrated=_integrated,
        )

    def to_py(self) -> None:
        raise ValueError("XmlFragment has no Python equivalent")

    def _get_or_insert(self, name: str, doc: Doc) -> Any:
        assert doc._txn is not None
        assert doc._txn._txn is not None
        return doc._doc.get_or_insert_xml_fragment(doc._txn._txn, name)

    def _init(self, value: list[XmlElement | str] | None) -> None:
        if value is None:
            return
        for obj in value:
            self.children.append(obj)


class XmlElement(_XmlFragmentTraitMixin, _XmlTraitMixin):
    _prelim: tuple[str, list[tuple[str, str]], list[str | XmlElement | XmlText]] | None
    _integrated: _XmlElement | None

    def __init__(
        self,
        tag: str | None = None,
        attributes: dict[str, str] | Iterable[tuple[str, str]] | None = None,
        contents: Iterable[XmlFragment | XmlElement | XmlText] | None = None,
        *,
        _doc: Doc | None = None,
        _integrated: _XmlElement | None = None,
    ) -> None:
        """
        Creates an XML element.

        Args:
            tag: The tag of the element (required).
            attributes: The optional attributes of the element.
            contents: The optional contents of the element.
        """
        if _integrated is not None:
            super().__init__(init=None, _doc=_doc, _integrated=_integrated)
            return

        if tag is None:
            raise ValueError("XmlElement: tag is required")

        if isinstance(attributes, dict):
            init_attrs = list(attributes.items())
        elif attributes is not None:
            init_attrs = list(attributes)
        else:
            init_attrs = []

        super().__init__(
            init=(
                tag,
                init_attrs,
                list(contents) if contents is not None else [],
            )
        )

    def to_py(self) -> None:
        raise ValueError("XmlElement has no Python equivalent")

    def _get_or_insert(self, name: str, doc: Doc) -> Any:
        raise ValueError("Cannot get an XmlElement from a doc, get an XmlFragment instead")

    def _init(
        self, value: tuple[str, list[tuple[str, str]], list[str | XmlElement | XmlText]] | None
    ):
        assert value is not None
        _, attrs, contents = value
        with self.doc.transaction():
            for k, v in attrs:
                self.attributes[k] = v
            for child in contents:
                self.children.append(child)

    @property
    def tag(self) -> str | None:
        """The element's tag, if any."""
        return self.integrated.tag()


class XmlText(_XmlTraitMixin):
    """
    A piece of text in an XML tree.

    This is similar to a [Text][pycrdt.Text], but instead of existing in a [Doc][pycrdt.Doc] on its
    own, it is a child of [XmlElement][pycrdt.XmlElement] or [XmlFragment][pycrdt.XmlFragment].
    """

    _prelim: str
    _integrated: _XmlText | None

    def __init__(
        self,
        init: str = "",
        *,
        _doc: Doc | None = None,
        _integrated: _XmlText | None = None,
    ) -> None:
        super().__init__(
            init=init,
            _doc=_doc,
            _integrated=_integrated,
        )

    def _get_or_insert(self, _name: str, _doc: Doc) -> Any:
        raise ValueError("Cannot get an XmlText from a doc - get an XmlFragment instead.")

    def to_py(self) -> str:
        if self._integrated is None:
            return self._prelim
        return str(self)

    def _init(self, value: str | None) -> None:  # pragma: no cover
        assert value is not None
        with self.doc.transaction() as txn:
            self.integrated.insert(txn._txn, 0, value)

    def __len__(self) -> int:
        with self.doc.transaction() as txn:
            return self.integrated.len(txn._txn)

    def __iadd__(self, value: str) -> XmlText:
        with self.doc.transaction():
            self.insert(len(self), value)
        return self

    def insert(self, index: int, value: str, attrs: Mapping[str, Any] | None = None) -> None:
        """
        Inserts text at a given index, with optional attributes.

        Args:
            index: The index at which to insert the text.
            value: The text to insert.
            attrs: The optional attributes.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            self.integrated.insert(
                txn._txn, index, value, iter(attrs.items()) if attrs is not None else iter([])
            )

    def insert_embed(self, index: int, value: Any, attrs: dict[str, Any] | None = None) -> None:
        """
        Insert an embed at a given index in the text, with optional attributes.

        Args:
            index: The index at which to insert the embed.
            value: The embed to insert.
            attrs: The optional attributes.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            _attrs = iter(attrs.items()) if attrs is not None else None
            if isinstance(value, BaseType):
                # shared type
                assert txn._txn is not None
                self._do_and_integrate("insert", value, txn._txn, index, _attrs)
            else:
                # primitive type
                self.integrated.insert_embed(txn._txn, index, value, _attrs)

    def format(self, start: int, stop: int, attrs: dict[str, Any]) -> None:
        """
        Formats existing text with attributes.

        Args:
            start: The index at which to start applying the attributes (included).
            stop: The index at which to stop applying the attributes (excluded).
            attrs: The attributes to apply.
        """
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            start, stop = _check_slice(self, slice(start, stop))
            length = stop - start
            if length > 0:
                self.integrated.format(txn._txn, start, length, iter(attrs.items()))

    def diff(self) -> list[tuple[Any, dict[str, Any] | None]]:
        """
        Returns:
            A list of formatted chunks that the current text corresponds to.
                Each list item is a tuple containing the chunk's contents and formatting attributes.
                The contents is usually the text as a string, but may be other data for embedded
                objects.
        """
        with self.doc.transaction() as txn:
            return self.integrated.diff(txn._txn)

    def __delitem__(self, key: int | slice) -> None:
        with self.doc.transaction() as txn:
            self._forbid_read_transaction(txn)
            if isinstance(key, int):
                self.integrated.remove_range(txn._txn, key, 1)
            elif isinstance(key, slice):
                start, stop = _check_slice(self, key)
                length = stop - start
                if length > 0:
                    self.integrated.remove_range(txn._txn, start, length)
            else:
                raise TypeError(f"Index not supported: {key}")

    def clear(self) -> None:
        """Removes the entire range of characters."""
        del self[:]


class XmlEvent(BaseEvent):
    __slots__ = ["children_changed", "target", "path", "delta", "keys", "transaction"]


class XmlAttributesView:
    """
    A list-like view into an [XmlFragment][pycrdt.XmlFragment] or [XmlElement][pycrdt.XmlElement]'s
    child nodes.

    Supports `len`, `in`, and getting, setting, and deleting by index. Iteration will iterate over
    key/value tuples.
    """

    inner: _XmlTraitMixin

    def __init__(self, inner: _XmlTraitMixin) -> None:
        self.inner = inner

    def get(self, key: str) -> Any | None:
        """
        Args:
            key: The name of the attribute to get.

        Returns:
            The value of the attribute, or `None` if there is no attribute with the given name.
        """
        with self.inner.doc.transaction() as txn:
            v = self.inner.integrated.attribute(txn._txn, key)
            if v is None:
                return None
            return v

    def __getitem__(self, key: str) -> Any:
        """
        Args:
            key: The name of the attribute to get.

        Raises:
            KeyError: Attribute does not exist.

        Returns:
            The attribute's value.
        """
        v = self.get(key)
        if v is None:
            raise KeyError(key)
        return v

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Args:
            key: The name of the attribute to set.
            value: The value of the attribute.
        """
        with self.inner.doc.transaction() as txn:
            self.inner._forbid_read_transaction(txn)
            self.inner.integrated.insert_attribute(txn._txn, key, value)

    def __delitem__(self, key: str) -> None:
        """
        Args:
            key: The value of the attribute to delete.
        """
        with self.inner.doc.transaction() as txn:
            self.inner._forbid_read_transaction(txn)
            self.inner.integrated.remove_attribute(txn._txn, key)

    def __contains__(self, key: str) -> bool:
        """
        Args:
            key: The name of the attribute to check.

        Returns:
            `True` if the attribute with the given name exists.
        """
        return self.get(key) is not None

    def __len__(self) -> int:
        """
        Returns:
            The number of attributes.
        """
        with self.inner.doc.transaction() as txn:
            return len(self.inner.integrated.attributes(txn._txn))

    def __iter__(self) -> Iterable[tuple[str, Any]]:
        """
        Returns:
            An iterable over each attribute, as key/value tuples.
        """
        with self.inner.doc.transaction() as txn:
            return iter(self.inner.integrated.attributes(txn._txn))


class XmlChildrenView:
    """
    A list-like view into an [XmlFragment][pycrdt.XmlFragment] or [XmlElement][pycrdt.XmlElement]'s
    child nodes.

    Supports `iter`, `len`, and getting, setting, and deleting by index.
    """

    inner: _XmlFragmentTraitMixin

    def __init__(self, inner: _XmlFragmentTraitMixin) -> None:
        self.inner = inner

    def __len__(self) -> int:
        """
        Returns:
            The number of child nodes.
        """
        with self.inner.doc.transaction() as txn:
            return self.inner.integrated.len(txn._txn)

    def __getitem__(self, index: int) -> XmlElement | XmlFragment | XmlText:
        """
        Args:
            index: The index of the child to get.

        Raises:
            IndexError: Index out of bounds.

        Returns:
            The child at the given index.
        """
        with self.inner.doc.transaction() as txn:
            if index >= len(self):
                raise IndexError(index)
            return _integrated_to_wrapper(
                self.inner.doc, self.inner.integrated.get(txn._txn, index)
            )

    def __delitem__(self, key: int | slice) -> None:
        """
        Args:
            key: The child index (`int`) or children `slice` to remove.
        """
        with self.inner.doc.transaction() as txn:
            self.inner._forbid_read_transaction(txn)
            if isinstance(key, int):
                self.inner.integrated.remove_range(txn._txn, key, 1)
            elif isinstance(key, slice):
                start, stop = _check_slice(self, key)
                length = stop - start
                if length > 0:
                    self.inner.integrated.remove_range(txn._txn, start, length)
            else:
                raise TypeError(f"Index not supported: {key}")

    def __setitem__(self, key: int, value: str | XmlText | XmlElement):
        """
        Replaces a child. Equivalent to deleting the index and then inserting the new value.

        Args:
            key: The index of the child to replace.
            value: The new value at the index.
        """
        with self.inner.doc.transaction():
            del self[key]
            self.insert(key, value)

    def __iter__(self) -> Iterator[XmlText | XmlElement | XmlFragment]:
        """
        Returns:
            An iterable over child nodes.
        """
        with self.inner.doc.transaction():
            children = [self[i] for i in range(len(self))]
        return iter(children)

    @overload
    def insert(self, index: int, element: str | XmlText) -> XmlText: ...
    @overload
    def insert(self, index: int, element: XmlElement) -> XmlElement: ...

    def insert(self, index: int, element: str | XmlText | XmlElement) -> XmlText | XmlElement:
        """
        Inserts a new node into the element's or fragment's children at the specified index.

        Passing in a `str` will convert it to an `XmlText`. Returns the passed in element, which
        will now be integrated into the tree.

        Args:
            index: The index at which to insert the element.
            element: The element to insert.

        Returns:
            The inserted element.
        """
        with self.inner.doc.transaction() as txn:
            self.inner._forbid_read_transaction(txn)
            if index > len(self):
                raise IndexError(index)
            if isinstance(element, str):
                integrated = self.inner.integrated.insert_str(txn._txn, index, element)
                return XmlText(_doc=self.inner.doc, _integrated=integrated)
            elif isinstance(element, XmlText):
                if element._integrated is not None:
                    raise ValueError("Cannot insert an integrated XmlText")
                integrated = self.inner.integrated.insert_str(txn._txn, index, element.prelim)
                element._integrate(self.inner.doc, integrated)
                return element
            elif isinstance(element, XmlElement):
                if element._integrated is not None:
                    raise ValueError("Cannot insert an integrated XmlElement")
                prelim = element.prelim
                integrated = self.inner.integrated.insert_element_prelim(txn._txn, index, prelim[0])
                element._integrate(self.inner.doc, integrated)
                element._init(prelim)
                return element
            else:
                raise TypeError("Cannot add value to XML: " + repr(element))

    @overload
    def append(self, element: str | XmlText) -> XmlText: ...
    @overload
    def append(self, element: XmlElement) -> XmlElement: ...

    def append(self, element: str | XmlText | XmlElement) -> XmlText | XmlElement:
        """
        Appends a new node to the end of the element's or fragment's children.

        Equivalent to `insert` at index `len(self)`.

        Args:
            element: The element to append.

        Returns:
            The appended element.
        """
        return self.insert(len(self), element)


base_types[_XmlFragment] = XmlFragment
base_types[_XmlElement] = XmlElement
base_types[_XmlText] = XmlText
event_types[_XmlEvent] = XmlEvent



from __future__ import annotations

from time import time_ns
from typing import TYPE_CHECKING, Any, Callable

from ._base import BaseType
from ._pycrdt import (
    StackItem,
)
from ._pycrdt import (
    UndoManager as _UndoManager,
)
from ._transaction import hash_origin

if TYPE_CHECKING:
    from ._doc import Doc


def timestamp() -> int:
    return time_ns() // 1_000_000


class UndoManager:
    """
    The undo manager allows to perform undo/redo operations on shared types.
    It can be initialized either with a [Doc][pycrdt.Doc] or with scopes.
    Scopes are a list of shared types integrated in a document.
    If initialized with a `Doc`, scopes can later be expanded.
    Changes can be undone/redone by batches using time intervals.
    It is possible to include/exclude changes by transaction origin in undo/redo operations.
    """

    def __init__(
        self,
        *,
        doc: Doc | None = None,
        scopes: list[BaseType] = [],
        capture_timeout_millis: int = 500,
        timestamp: Callable[[], int] = timestamp,
        undo_stack: list[StackItem] | None = None,
        redo_stack: list[StackItem] | None = None,
    ) -> None:
        """
        Args:
            doc: The document the undo manager will work with.
            scopes: A list of shared types the undo manager will work with.
            capture_timeout_millis: A time interval for grouping changes that will be undone/redone.
            timestamp: A function that returns a timestamp as an integer number of milli-seconds.
            undo_stack: Pre-filled undo stack items.
            redo_stack: Pre-filled redo stack items.

        Raises:
            RuntimeError: UndoManager must be created with doc or scopes.
        """
        if doc is None:
            if not scopes:
                raise RuntimeError("UndoManager must be created with doc or scopes")
            doc = scopes[0].doc
        elif scopes:
            raise RuntimeError("UndoManager must be created with doc or scopes")
        self._undo_manager = _UndoManager(
            doc._doc,
            capture_timeout_millis,
            timestamp,
            undo_stack or [],
            redo_stack or [],
        )
        for scope in scopes:
            self.expand_scope(scope)

    def expand_scope(self, scope: BaseType) -> None:
        """
        Expands the scope of shared types for this undo manager.

        Args:
            scope: The shared type to include.
        """
        method = getattr(self._undo_manager, f"expand_scope_{scope.type_name}")
        method(scope._integrated)

    def include_origin(self, origin: Any) -> None:
        """
        Extends the list of transactions origin tracked by this undo manager.

        Args:
            origin: The origin to include.
        """
        self._undo_manager.include_origin(hash_origin(origin))

    def exclude_origin(self, origin: Any) -> None:
        """
        Removes a transaction origin from the list of origins tracked by this undo manager.

        Args:
            origin: The origin to exclude.
        """
        self._undo_manager.exclude_origin(hash_origin(origin))

    def can_undo(self) -> bool:
        """
        Returns:
            True if there are changes to undo.
        """
        return self._undo_manager.can_undo()

    def undo(self) -> bool:
        """
        Perform an undo operation.

        Returns:
            True if some changes were undone.
        """
        return self._undo_manager.undo()

    def can_redo(self) -> bool:
        """
        Returns:
            True if there are changes to redo.
        """
        return self._undo_manager.can_redo()

    def redo(self) -> bool:
        """
        Perform a redo operation.

        Returns:
            True if some changes were redone.
        """
        return self._undo_manager.redo()

    def clear(self) -> None:
        """
        Clears all [StackItem][pycrdt.StackItem]s stored in this undo manager,
        effectively resetting its state.
        """
        self._undo_manager.clear()

    @property
    def undo_stack(self) -> list[StackItem]:
        """The list of undoable actions."""
        return self._undo_manager.undo_stack()

    @property
    def redo_stack(self) -> list[StackItem]:
        """The list of redoable actions."""
        return self._undo_manager.redo_stack()



from ._pycrdt import get_state as _get_state
from ._pycrdt import get_update as _get_update
from ._pycrdt import merge_updates as _merge_updates


def get_state(update: bytes) -> bytes:
    """
    Returns a state from an update.

    Args:
        update: The update from which to get the state.

    Returns:
        The state corresponding to the update.
    """
    return _get_state(update)


def get_update(update: bytes, state: bytes) -> bytes:
    """
    Returns an update consisting of all changes from a given update which have not
    been seen in the given state.

    Args:
        update: The update from which to get all missing changes in the given state.
        state: The state from which to get missing changes that are in the given update.

    Returns:
        The changes from the given update not present in the given state.
    """
    return _get_update(update, state)


def merge_updates(*updates: bytes) -> bytes:
    """
    Returns an update consisting of a combination of all given updates.

    Args:
        updates: The updates to merge.

    Returns:
        The merged updates.
    """
    return _merge_updates(updates)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ._pycrdt import Snapshot as _Snapshot

if TYPE_CHECKING:
    from ._doc import Doc


@dataclass
class Snapshot:
    """
    A snapshot of a document's state at a given point in time.
    Can be encoded to bytes for storage or transmission, and decoded back.
    """

    _snapshot: _Snapshot

    @classmethod
    def from_doc(cls, doc: "Doc") -> "Snapshot":
        """
        Create a snapshot from a document.

        Args:
            doc: The document to snapshot.
        Returns:
            The snapshot of the document's current state.
        """
        snap = _Snapshot.from_doc(doc._doc)
        return cls(snap)

    @classmethod
    def decode(cls, data: bytes) -> Snapshot:
        """
        Decode a snapshot from its binary representation.

        Args:
            data: The bytes to decode into a snapshot.
        Returns:
            The decoded snapshot.
        Raises:
            ValueError: If the bytes are not a valid snapshot.
        """
        snap = _Snapshot.decode(data)
        return cls(snap)

    def encode(self) -> bytes:
        """
        Encode the snapshot to its binary representation.

        Returns:
            The bytes representing the snapshot.
        """
        return self._snapshot.encode()



from __future__ import annotations

import copy
import json
from time import time
from typing import Any, Callable, Literal, cast
from uuid import uuid4

from anyio import TASK_STATUS_IGNORED, create_task_group, sleep
from anyio.abc import TaskGroup, TaskStatus

from ._doc import Doc
from ._sync import Decoder, Encoder, read_message


class Awareness:
    client_id: int
    _meta: dict[int, dict[str, Any]]
    _states: dict[int, dict[str, Any]]
    _subscriptions: dict[str, Callable[[str, tuple[dict[str, Any], Any]], None]]
    _task_group: TaskGroup | None

    def __init__(self, ydoc: Doc, *, outdated_timeout: int = 30000) -> None:
        """
        Args:
            ydoc: The [Doc][pycrdt.Doc] to associate the awareness with.
            outdated_timeout: The timeout (in milliseconds) to consider a client gone.
        """
        self.client_id = ydoc.client_id
        self._outdated_timeout = outdated_timeout
        self._meta = {}
        self._states = {}
        self._subscriptions = {}
        self._task_group = None
        self.set_local_state({})

    @property
    def meta(self) -> dict[int, dict[str, Any]]:
        """The clients' metadata."""
        return self._meta

    @property
    def states(self) -> dict[int, dict[str, Any]]:
        """The client states."""
        return self._states

    def _emit(
        self,
        topic: Literal["change", "update"],
        added: list[int],
        updated: list[int],
        removed: list[int],
        origin: Any,
    ):
        for callback in self._subscriptions.values():
            callback(topic, ({"added": added, "updated": updated, "removed": removed}, origin))

    def _get_time(self) -> int:
        return int(time() * 1000)

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        """
        Starts updating the awareness periodically.
        """
        if self._task_group is not None:
            raise RuntimeError("Awareness already started")

        async with create_task_group() as tg:
            self._task_group = tg
            task_status.started()
            tg.start_soon(self._start)

    async def _start(self) -> None:
        while True:
            await sleep(self._outdated_timeout / 1000 / 10)
            now = self._get_time()
            if (
                self.get_local_state() is not None
                and self._outdated_timeout / 2 <= now - self._meta[self.client_id]["lastUpdated"]
            ):
                # renew local clock
                self.set_local_state(self.get_local_state())
            remove: list[int] = []
            for client_id, meta in self._meta.items():
                if (
                    client_id != self.client_id
                    and self._outdated_timeout <= now - meta["lastUpdated"]
                    and client_id in self._states
                ):
                    remove.append(client_id)
            if remove:
                self.remove_awareness_states(remove, "timeout")

    async def stop(self) -> None:
        """
        Stops updating the awareness periodically.
        """
        if self._task_group is None:
            raise RuntimeError("Awareness not started")
        self._task_group.cancel_scope.cancel()
        self._task_group = None

    def get_local_state(self) -> dict[str, Any] | None:
        """
        Returns:
            The local state, if any.
        """
        return self._states.get(self.client_id)

    def set_local_state(self, state: dict[str, Any] | None) -> None:
        """
        Updates the local state and meta, and sends the changes to subscribers.

        Args:
            state: The new local state, if any.
        """
        client_id = self.client_id
        curr_local_meta = self._meta.get(client_id)
        clock = 0 if curr_local_meta is None else curr_local_meta["clock"] + 1
        prev_state = self._states.get(client_id)
        if prev_state is not None:
            prev_state = copy.deepcopy(prev_state)
        if state is None:
            if client_id in self._states:
                del self._states[client_id]
        else:
            self._states[client_id] = state
        timestamp = self._get_time()
        self._meta[client_id] = {"clock": clock, "lastUpdated": timestamp}
        added = []
        updated = []
        filtered_updated = []
        removed = []
        if state is None:
            removed.append(client_id)
        elif prev_state is None:
            if state is not None:
                added.append(client_id)
        else:
            updated.append(client_id)
            if prev_state != state:
                filtered_updated.append(client_id)
        if added or filtered_updated or removed:
            self._emit("change", added, filtered_updated, removed, "local")
        self._emit("update", added, updated, removed, "local")

    def set_local_state_field(self, field: str, value: Any) -> None:
        """
        Sets a local state field.

        Args:
            field: The field of the local state to set.
            value: The value associated with the field.
        """
        state = self.get_local_state()
        if state is not None:
            state = copy.deepcopy(state)
            state[field] = value
            self.set_local_state(state)

    def remove_awareness_states(self, client_ids: list[int], origin: Any) -> None:
        """
        Removes awareness states for clients given by their IDs.

        Args:
            client_ids: The list of client IDs for which to remove the awareness states.
            origin: The origin of the update.
        """
        removed = []
        for client_id in client_ids:
            if client_id in self._states:
                del self._states[client_id]
                if client_id == self.client_id:
                    cur_meta = self._meta[client_id]
                    self._meta[client_id] = {
                        "clock": cur_meta["clock"] + 1,
                        "lastUpdated": self._get_time(),
                    }
                removed.append(client_id)
        if removed:
            self._emit("change", [], [], removed, origin)
            self._emit("update", [], [], removed, origin)

    def encode_awareness_update(self, client_ids: list[int]) -> bytes:
        """
        Creates an encoded awareness update of the clients given by their IDs.

        Args:
            client_ids: The list of client IDs for which to create an update.

        Returns:
            The encoded awareness update.
        """
        encoder = Encoder()
        encoder.write_var_uint(len(client_ids))
        for client_id in client_ids:
            state = self._states.get(client_id)
            clock = cast(int, self._meta.get(client_id, {}).get("clock"))
            encoder.write_var_uint(client_id)
            encoder.write_var_uint(clock)
            encoder.write_var_string(json.dumps(state, separators=(",", ":")))
        return encoder.to_bytes()

    def apply_awareness_update(self, update: bytes, origin: Any) -> None:
        """
        Applies the binary update and notifies subscribers with changes.

        Args:
            update: The binary update.
            origin: The origin of the update.
        """
        decoder = Decoder(update)
        timestamp = self._get_time()
        added = []
        updated = []
        filtered_updated = []
        removed = []
        length = decoder.read_var_uint()
        for _ in range(length):
            client_id = decoder.read_var_uint()
            clock = decoder.read_var_uint()
            state_str = decoder.read_var_string()
            state = None if not state_str else json.loads(state_str)
            client_meta = self._meta.get(client_id)
            prev_state = self._states.get(client_id)
            curr_clock = 0 if client_meta is None else client_meta["clock"]
            if curr_clock < clock or (
                curr_clock == clock and state is None and client_id in self._states
            ):
                if state is None:
                    # Never let a remote client remove this local state.
                    if client_id == self.client_id and self.get_local_state() is not None:
                        # Remote client removed the local state. Do not remove state.
                        # Broadcast a message indicating that this client still exists by increasing
                        # the clock.
                        clock += 1
                    else:
                        if client_id in self._states:
                            del self._states[client_id]
                else:
                    self._states[client_id] = state
                self._meta[client_id] = {
                    "clock": clock,
                    "lastUpdated": timestamp,
                }
                if client_meta is None and state is not None:
                    added.append(client_id)
                elif client_meta is not None and state is None:
                    removed.append(client_id)
                elif state is not None:
                    if state != prev_state:
                        filtered_updated.append(client_id)
                    updated.append(client_id)
        if added or filtered_updated or removed:
            self._emit("change", added, filtered_updated, removed, origin)
        if added or updated or removed:
            self._emit("update", added, updated, removed, origin)

    def observe(self, callback: Callable[[str, tuple[dict[str, Any], Any]], None]) -> str:
        """
        Registers the given callback to awareness changes.

        Args:
            callback: The callback to call with the awareness changes.

        Returns:
            The subscription ID that can be used to unobserve.
        """
        id = str(uuid4())
        self._subscriptions[id] = callback
        return id

    def unobserve(self, id: str) -> None:
        """
        Unregisters the given subscription ID from awareness changes.

        Args:
            id: The subscription ID to unregister.
        """
        del self._subscriptions[id]


def is_awareness_disconnect_message(message: bytes) -> bool:
    """
    Check if the message is null, which means that it is a disconnection message
    from the client.

    Args:
        message: The message received from the client.

    Returns:
        Whether the message is a disconnection message or not.
    """
    decoder = Decoder(read_message(message))
    length = decoder.read_var_uint()
    # A disconnection message should be a single message
    if length == 1:
        # Remove client_id and clock information from message (not used)
        for _ in range(2):
            decoder.read_var_uint()
        state = decoder.read_var_string()
        if state == "null":
            return True
    return False
